import torch
import torch.nn as nn
from timm.models.layers.helpers import make_divisible

from .attention_modules import Attention
from .base_modules import Activation, Conv, Norm

__all__ = [
    "BasicConv",
    "BottleneckConv",
    "DepthWiseSeparableConv",
    "InvertedBottleneckConv",
    "FusedMobileInvertedConv",
    "HoverNetDenseConv",
    "BasicConvOld",
]


class BasicConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size=3,
        groups: int = 1,
        bias: bool = False,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Conv-block (basic) parent class.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            groups : int, default=1
                Number of groups the kernels are divided into. If `groups == 1`
                normal convolution is applied. If `groups = in_channels`
                depthwise convolution is applied.
            bias : bool, default=False,
                Include bias term in the convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()
        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        # set norm channel number for preactivation or normal
        norm_channels = in_channels if preactivate else self.out_channels

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0

        self.conv = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
            padding=padding,
            bias=bias,
        )

        self.norm = Norm(normalization, num_features=norm_channels)
        self.act = Activation(activation)

        # set attention channels
        att_channels = in_channels if preattend else self.out_channels
        self.att = Attention(attention, in_channels=att_channels)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        x = self.conv(x)
        x = self.norm(x)
        x = self.act(x)

        if not self.preattend:
            x = self.att(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        x = self.norm(x)
        x = self.act(x)
        x = self.conv(x)

        if not self.preattend:
            x = self.att(x)

        return x


class BottleneckConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expand_ratio: float = 4.0,
        base_width: int = 64,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size: int = 3,
        groups: int = 1,
        bias: bool = False,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Bottleneck conv block parent-class.

        Res-Net: Deep residual learning for image recognition:
            - https://arxiv.org/abs/1512.03385

        Preact-ResNet: Identity Mappings in Deep Residual Networks:
            - https://arxiv.org/abs/1603.05027

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            expand_ratio : float, default=4.0
                The ratio of channel expansion in the bottleneck.
            base_width : int, default=64
                The minimum width for the conv x channels in this block.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            groups : int, default=1
                Number of groups the kernels are divided into. If `groups == 1`
                normal convolution is applied. If `groups = in_channels`
                depthwise convolution is applied.
            bias : bool, default=False,
                Include bias term in the convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()
        self.conv_choice = convolution
        self.expansion = int(expand_ratio)
        self.preattend = preattend
        self.preactivate = preactivate

        width = int(out_channels * (base_width / 64.0)) * groups
        self.out_channels = out_channels * self.expansion

        # set attention channels
        att_channels = in_channels if preattend else self.out_channels
        self.att = Attention(attention, in_channels=att_channels)

        self.conv1 = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            out_channels=width,
            kernel_size=1,
            bias=bias,
            padding=0,
        )

        norm_channels = in_channels if preactivate else width
        self.norm1 = Norm(normalization, num_features=norm_channels)
        self.act1 = Activation(activation)

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0
        self.conv2 = Conv(
            name=self.conv_choice,
            in_channels=width,
            out_channels=width,
            kernel_size=kernel_size,
            groups=groups,
            padding=padding,
            bias=bias,
        )

        self.norm2 = Norm(normalization, num_features=width)
        self.act2 = Activation(activation)

        self.conv3 = Conv(
            name=self.conv_choice,
            in_channels=width,
            bias=bias,
            out_channels=self.out_channels,
            kernel_size=1,
            padding=0,
        )

        norm_channels = width if preactivate else self.out_channels
        self.norm3 = Norm(normalization, num_features=norm_channels)

        self.act3 = Activation(activation)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act1(x)

        x = self.conv2(x)
        x = self.norm2(x)
        x = self.act2(x)

        x = self.conv3(x)
        x = self.norm3(x)

        x = self.act3(x)

        if not self.preattend:
            x = self.att(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        x = self.norm1(x)
        x = self.act1(x)
        x = self.conv1(x)

        x = self.norm2(x)
        x = self.act2(x)
        x = self.conv2(x)

        x = self.norm3(x)
        x = self.conv3(x)

        x = self.act3(x)

        if not self.preattend:
            x = self.att(x)

        return x


class DepthWiseSeparableConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size: int = 3,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Depthwise separable conv block parent class.

        MobileNets:
        Efficient Convolutional Neural Networks for Mobile Vision Applications:
            - https://arxiv.org/abs/1704.04861

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()
        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0

        self.depth_conv = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            groups=in_channels,
            padding=padding,
        )

        self.norm1 = Norm(normalization, num_features=in_channels)
        self.act1 = Activation(activation)
        self.att = Attention(attention, in_channels=in_channels)

        self.ch_pool = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            padding=0,
            out_channels=self.out_channels,
            kernel_size=1,
            bias=False,
        )
        norm_channels = in_channels if preactivate else self.out_channels
        self.norm2 = Norm(normalization, num_features=norm_channels)
        self.act2 = Activation(activation)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        x = self.depth_conv(x)
        x = self.norm1(x)
        x = self.act1(x)

        if not self.preattend:
            x = self.att(x)

        # pointwise channel pool
        x = self.ch_pool(x)
        x = self.norm2(x)
        x = self.act2(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        x = self.norm1(x)
        x = self.act1(x)
        x = self.depth_conv(x)

        if not self.preattend:
            x = self.att(x)

        # pointwise channel pool
        x = self.norm2(x)
        x = self.act2(x)
        x = self.ch_pool(x)

        return x


class InvertedBottleneckConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expand_ratio: float = 4.0,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size: int = 3,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Mobile inverted bottleneck conv parent-class.

        MobileNetV2: Inverted Residuals and Linear Bottlenecks:
            - https://arxiv.org/abs/1801.04381

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            expand_ratio : float, default=4.0
                The ratio of channel expansion in the bottleneck.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()

        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        mid_channels = make_divisible(in_channels * expand_ratio)

        self.ch_pool = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            padding=0,
            out_channels=mid_channels,
            kernel_size=1,
            bias=False,
        )

        norm_channels = in_channels if preactivate else mid_channels
        self.norm1 = Norm(normalization, num_features=norm_channels)
        self.act1 = Activation(activation)

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0
        self.depth_conv = Conv(
            name=self.conv_choice,
            in_channels=mid_channels,
            out_channels=mid_channels,
            kernel_size=kernel_size,
            groups=mid_channels,
            padding=padding,
            bias=False,
        )
        self.norm2 = Norm(normalization, num_features=mid_channels)
        self.act2 = Activation(activation)

        # set attention
        att_channels = in_channels if preattend else mid_channels
        self.att = Attention(attention, in_channels=att_channels, squeeze_ratio=0.04)

        self.proj_conv = Conv(
            name=self.conv_choice,
            in_channels=mid_channels,
            bias=False,
            out_channels=self.out_channels,
            kernel_size=1,
            padding=0,
        )

        norm_channels = mid_channels if preactivate else self.out_channels
        self.norm3 = Norm(normalization, num_features=norm_channels)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        # pointwise channel pooling conv
        x = self.ch_pool(x)
        x = self.norm1(x)
        x = self.act1(x)

        # depthwise conv
        x = self.depth_conv(x)
        x = self.norm2(x)
        x = self.act2(x)

        if not self.preattend:
            x = self.att(x)

        # Pointwise linear projection
        x = self.proj_conv(x)
        x = self.norm3(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        x = self.norm1(x)
        x = self.act1(x)
        x = self.ch_pool(x)

        # depthwise conv
        x = self.norm2(x)
        x = self.act2(x)
        x = self.depth_conv(x)

        if not self.preattend:
            x = self.att(x)

        # Pointwise linear projection
        x = self.norm3(x)
        x = self.proj_conv(x)

        return x


class FusedMobileInvertedConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expand_ratio: float = 4.0,
        kernel_size: int = 3,
        stride: int = 1,
        same_padding: bool = True,
        convolution: str = "conv",
        preactivate: bool = False,
        normalization: str = "bn",
        activation: str = "relu",
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Fused mobile inverted conv block parent-class.

        Efficientnet-edgetpu: Creating accelerator-optimized neural networks with automl
            - https://ai.googleblog.com/2019/08/efficientnet-edgetpu-creating.html

        EfficientNetV2: Smaller Models and Faster Training
            - https://arxiv.org/abs/2104.00298

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            expand_ratio : float, default=4.0
                The ratio of channel expansion in the bottleneck.
            kernel_size : int, default=3
                The size of the convolution kernel.
            stride : int, default=1
                The stride size.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()
        assert stride in (1, 2)

        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        mid_channels = make_divisible(in_channels * expand_ratio)

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0
        self.conv1 = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            padding=padding,
            out_channels=mid_channels,
            kernel_size=kernel_size,
            bias=False,
            stride=stride,
        )

        norm_channels = in_channels if preactivate else mid_channels
        self.norm1 = Norm(normalization, num_features=norm_channels)

        att_channels = in_channels if preattend else mid_channels
        self.att = Attention(attention, in_channels=att_channels, squeeze_ratio=0.04)

        self.proj_conv = Conv(
            name=self.conv_choice,
            in_channels=mid_channels,
            bias=False,
            out_channels=self.out_channels,
            kernel_size=1,
            padding=0,
        )

        norm_channels = mid_channels if preactivate else self.out_channels
        self.norm2 = Norm(normalization, num_features=norm_channels)
        self.act = Activation(activation)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        # pointwise channel pooling conv
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.act(x)

        if not self.preattend:
            x = self.att(x)

        # Pointwise linear projection
        x = self.proj_conv(x)
        x = self.norm2(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        # pointwise channel pooling conv
        x = self.norm1(x)
        x = self.act(x)
        x = self.conv1(x)

        if not self.preattend:
            x = self.att(x)

        # Pointwise linear projection
        x = self.norm2(x)
        x = self.proj_conv(x)

        return x


class HoverNetDenseConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        squeeze_ratio: float = 4.0,
        groups: int = 4,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size: int = 3,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Dense block of the HoVer-Net.

        HoVer-Net:
        - https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            squeeze_ratio : float, default=4.0
                The ratio of channel expansion in the bottleneck.
            groups : int, default=1
                Number of groups the kernels are divided into. If `groups == 1`
                normal convolution is applied. If `groups = in_channels`
                depthwise convolution is applied.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()

        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        mid_channels = make_divisible(in_channels / squeeze_ratio)
        self.ch_pool = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            padding=0,
            out_channels=mid_channels,
            kernel_size=1,
            bias=False,
        )

        norm_channels = in_channels if preactivate else mid_channels
        self.norm1 = Norm(normalization, num_features=norm_channels)
        self.act1 = Activation(activation)

        padding = (kernel_size - 1) // 2 if same_padding else 0
        self.conv = Conv(
            name=self.conv_choice,
            in_channels=mid_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
            padding=padding,
            bias=False,
        )
        norm_channels = mid_channels if preactivate else out_channels
        self.norm2 = Norm(normalization, num_features=norm_channels)
        self.act2 = Activation(activation)

        # set attention
        att_channels = in_channels if preattend else mid_channels
        self.att = Attention(attention, in_channels=att_channels, squeeze_ratio=0.04)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        if self.preattend:
            x = self.att(x)

        x = self.ch_pool(x)
        x = self.norm1(x)
        x = self.act1(x)

        if not self.preattend:
            x = self.att(x)

        # pointwise channel pool
        x = self.conv(x)
        x = self.norm2(x)
        x = self.act2(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        if self.preattend:
            x = self.att(x)

        x = self.norm1(x)
        x = self.act1(x)
        x = self.ch_pool(x)

        if not self.preattend:
            x = self.att(x)

        # pointwise channel pool
        x = self.norm2(x)
        x = self.act2(x)
        x = self.conv(x)

        return x


class BasicConvOld(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        same_padding: bool = True,
        normalization: str = "bn",
        activation: str = "relu",
        convolution: str = "conv",
        preactivate: bool = False,
        kernel_size=3,
        groups: int = 1,
        bias: bool = False,
        attention: str = None,
        preattend: bool = False,
        **kwargs
    ) -> None:
        """Conv-block (basic) parent class.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            same_padding : bool, default=True
                if True, performs same-covolution.
            normalization : str, default="bn":
                Normalization method.
                One of: "bn", "bcn", "gn", "in", "ln", None
            activation : str, default="relu"
                Activation method.
                One of: "mish", "swish", "relu", "relu6", "rrelu", "selu",
                "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolution : str, default="conv"
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivate : bool, default=False
                If True, normalization will be applied before convolution.
            kernel_size : int, default=3
                The size of the convolution kernel.
            groups : int, default=1
                Number of groups the kernels are divided into. If `groups == 1`
                normal convolution is applied. If `groups = in_channels`
                depthwise convolution is applied.
            bias : bool, default=False,
                Include bias term in the convolution.
            attention : str, default=None
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattend : bool, default=False
                If True, Attention is applied at the beginning of forward pass.
        """
        super().__init__()
        self.conv_choice = convolution
        self.out_channels = out_channels
        self.preattend = preattend
        self.preactivate = preactivate

        # set norm channel number for preactivation or normal
        norm_channels = in_channels if preactivate else self.out_channels

        # set padding. Works if dilation or stride are not adjusted
        padding = (kernel_size - 1) // 2 if same_padding else 0

        self.conv = Conv(
            name=self.conv_choice,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
            padding=padding,
            bias=bias,
        )

        self.norm = Norm(normalization, num_features=norm_channels)
        self.act = Activation(activation)

        # set attention channels
        att_channels = in_channels if preattend else self.out_channels
        self.att = Attention(attention, in_channels=att_channels)

        self.downsample = None
        if in_channels != out_channels:
            self.downsample = nn.Sequential(
                Conv(
                    self.conv_choice,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    bias=False,
                    kernel_size=1,
                    padding=0,
                ),
                Norm(normalization, num_features=out_channels),
            )

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        identity = x
        if self.downsample is not None:
            identity = self.downsample(x)

        x = self.att(x)

        # residual
        x = self.conv(x)
        x = self.norm(x)

        x += identity
        x = self.act(x)

        return x

    def forward_features_preact(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass with pre-activation."""
        identity = x
        if self.downsample is not None:
            identity = self.downsample(x)

        # pre-attention
        x = self.att(x)

        # preact residual
        x = self.norm(x)
        x = self.act(x)
        x = self.conv(x)

        x += identity
        x = self.act(x)

        return x



from .base._multitask_unet import MultiTaskUnet
from .cellpose.cellpose import (
    CellPoseUnet,
    cellpose_base,
    cellpose_plus,
    omnipose_base,
    omnipose_plus,
)
from .hovernet.hovernet import (
    HoverNet,
    hovernet_base,
    hovernet_plus,
    hovernet_small,
    hovernet_small_plus,
)
from .stardist.stardist import (
    StarDistUnet,
    stardist_base,
    stardist_base_multiclass,
    stardist_plus,
)

__all__ = [
    "MultiTaskUnet",
    "HoverNet",
    "hovernet_base",
    "hovernet_plus",
    "hovernet_small",
    "hovernet_small_plus",
    "CellPoseUnet",
    "cellpose_base",
    "cellpose_plus",
    "omnipose_base",
    "omnipose_plus",
    "StarDistUnet",
    "stardist_base",
    "stardist_plus",
    "stardist_base_multiclass",
]



from typing import List, Tuple

import torch
import torch.nn as nn

from .conv_block import ConvBlock
from .misc_modules import ChannelPool

__all__ = ["ConvLayer"]


class ConvLayer(nn.ModuleDict):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        n_blocks: int = 2,
        layer_residual: bool = False,
        short_skip: str = "residual",
        style_channels: int = None,
        expand_ratios: Tuple[float, float] = (1.0, 1.0),
        block_types: Tuple[str, ...] = ("basic", "basic"),
        normalizations: Tuple[str, ...] = ("bn", "bn"),
        activations: Tuple[str, ...] = ("relu", "relu"),
        convolutions: Tuple[str, ...] = ("conv", "conv"),
        kernel_sizes: Tuple[int, ...] = (3, 3),
        groups: Tuple[int, ...] = (1, 1),
        biases: Tuple[bool, ...] = (True, True),
        preactivates: Tuple[bool, ...] = (False, False),
        attentions: Tuple[str, ...] = (None, None),
        preattends: Tuple[bool, ...] = (False, False),
        use_styles: Tuple[bool, ...] = (False, False),
        **kwargs,
    ) -> None:
        """Stack conv-blocks in a ModuleDict to compose a full layer.

        Optional:
            - add a style vector to the output at the end of each conv block (Cellpose)

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            out_channels : int
                Number of output channels.
            n_blocks : int, default=2
                Number of ConvBlocks used in this layer.
            layer_residual : bool, default=False
                Apply a layer level residual skip. I.e x + layer(x)
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
            short_skip : str, default="residual"
                The name of the short skip method. One of: "residual", "dense", "basic"
            expand_ratios : Tuple[float, ...], default=(1.0, 1.0):
                Expansion/Squeeze ratios for the out channels of each conv block.
            block_types : Tuple[str, ...], default=("basic", "basic")
                The name of the conv-blocks. Length of the tuple has to equal `n_blocks`
                One of: "basic". "mbconv", "fmbconv" "dws", "bottleneck".
            normalizations : Tuple[str, ...], default=("bn", "bn"):
                Normalization methods. One of: "bn", "bcn", "gn", "in", "ln", "lrn"
            activations : Tuple[str, ...], default=("relu", "relu")
                Activation methods. One of: "mish", "swish", "relu", "relu6", "rrelu",
                "selu", "celu", "gelu", "glu", "tanh", "sigmoid", "silu", "prelu",
                "leaky-relu", "elu", "hardshrink", "tanhshrink", "hardsigmoid"
            convolutions : Tuple[str, ...], default=("conv", "conv")
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            preactivates : Tuple[bool, ...], default=(False, False)
                Pre-activations flags for the conv-blocks.
            kernel_sizes : Tuple[int, ...], default=(3, 3)
                The size of the convolution kernels in each conv block.
            groups : int, default=(1, 1)
                Number of groups for the kernels in each convolution blocks.
            biases : bool, default=(True, True)
                Include bias terms in the convolution blocks.
            attentions : Tuple[str, ...], default=(None, None)
                Attention method. One of: "se", "scse", "gc", "eca", None
            preattends : Tuple[bool, ...], default=(False, False)
                If True, Attention is applied at the beginning of forward pass.
            use_styles : bool, default=(False, False)
                If True and `style_channels` is not None, adds a style vec to output.

        Raises
        ------
            ValueError:
                If lengths of the tuple arguments are not equal to `n_blocks`.
        """
        super().__init__()
        self.layer_residual = layer_residual
        self.short_skip = short_skip
        self.in_channels = in_channels

        illegal_args = [
            (k, a)
            for k, a in locals().items()
            if isinstance(a, tuple) and len(a) != n_blocks
        ]

        if illegal_args:
            raise ValueError(
                f"""
                All the tuple-arg lengths need to be equal to `n_blocks`={n_blocks}.
                Illegal args: {illegal_args}"""
            )

        blocks = list(range(n_blocks))
        for i in blocks:
            out = int(out_channels * expand_ratios[i])

            conv_block = ConvBlock(
                name=block_types[i],
                in_channels=in_channels,
                out_channels=out,
                style_channels=style_channels,
                short_skip=short_skip,
                kernel_size=kernel_sizes[i],
                groups=groups[i],
                bias=biases[i],
                normalization=normalizations[i],
                convolution=convolutions[i],
                activation=activations[i],
                attention=attentions[i],
                preactivate=preactivates[i],
                preattend=preattends[i],
                use_style=use_styles[i],
                **kwargs,
            )
            self.add_module(f"{short_skip}_{block_types[i]}_{i + 1}", conv_block)

            if short_skip == "dense":
                in_channels += conv_block.out_channels
            else:
                in_channels = conv_block.out_channels

        self.out_channels = conv_block.out_channels

        if short_skip == "dense":
            self.transition = ConvBlock(
                name="basic",
                in_channels=in_channels,
                short_skip="basic",
                out_channels=out_channels,
                same_padding=False,
                bias=False,
                kernel_size=1,
                convolution=conv_block.block.conv_choice,
                normalization=normalizations[-1],
                activation=activations[-1],
                preactivate=preactivates[-1],
            )
            self.out_channels = self.transition.out_channels

        self.downsample = None
        if layer_residual and self.in_channels != self.out_channels:
            self.downsample = ChannelPool(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                convolution=convolutions[-1],
                normalization=normalizations[-1],
            )

    def forward_features_dense(
        self, init_features: List[torch.Tensor], style: torch.Tensor = None
    ) -> torch.Tensor:
        """Dense forward pass."""
        features = [init_features]
        for name, conv_block in self.items():
            if name not in ("transition", "downsample"):
                new_features = conv_block(features, style)
                features.append(new_features)

        x = torch.cat(features, 1)
        x = self.transition(x)

        return x

    def forward_features(
        self, x: torch.Tensor, style: torch.Tensor = None
    ) -> torch.Tensor:
        """Regular forward pass."""
        for name, conv_block in self.items():
            if name != "downsample":
                x = conv_block(x, style)

        return x

    def forward(self, x: torch.Tensor, style: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of the conv-layer."""
        if self.layer_residual:
            identity = x
            if self.downsample is not None:
                identity = self.downsample(x)

        if self.short_skip == "dense":
            x = self.forward_features_dense(x, style)
        else:
            x = self.forward_features(x, style)

        if self.layer_residual:
            x = x + identity

        return x



from typing import Any, Dict, List

import torch.nn as nn


def adjust_optim_params(
    model: nn.Module, optim_params: Dict[str, Dict[str, Any]]
) -> List[Dict[str, Dict]]:
    """Adjust the learning parameters for optimizer.

    Parameters
    ----------
        model : nn.Module
            The encoder-decoder segmentation model.
        optim_params : Dict[str, Dict[str, Any]]
            optim paramas like learning rates, weight decays etc for diff parts of
            the network. E.g.
            {"encoder": {"weight_decay: 0.1, "lr":0.1}, "sem": {"lr": 0.1}}

    Returns
    -------
        List[Dict[str, Dict]]:
            a list of kwargs (str, Dict pairs) containing the model params.
    """
    params = list(model.named_parameters())

    adjust_params = []
    for name, parameters in params:
        opts = {}

        for block, block_params in optim_params.items():
            if block in name:
                for key, item in block_params.items():
                    opts[key] = item

        adjust_params.append({"params": parameters, **opts})

    return adjust_params



"""
Imported from DRFNS repo. Code style mods.

https://github.com/PeterJackNaylor/DRFNS/

MIT License

Copyright (c) 2018 PeterJackNaylor

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
import scipy.ndimage as ndi
import skimage.morphology as morph
import skimage.segmentation as segm
from skimage import img_as_ubyte

from cellseg_models_pytorch.utils import binarize, percentile_normalize99

__all__ = ["post_proc_drfns"]


def h_minima_reconstruction(inv_dist_map: np.ndarray, lamb: int = 7) -> np.ndarray:
    """Perform a H minimma reconstruction via an erosion method.

    Parameters
    ----------
        inv_dist_map : np.ndarray
            Inverse distance map. Shape: (H, W).
        lamb : int, default=7
            Intensity shift value lambda.

    Returns
    -------
        np.ndarray:
            H minima reconstruction from the inverse distance transform.
            Shape: (H, W).
    """

    def making_top_mask(x: np.ndarray, lamb: int = lamb) -> int:
        return min(255, x + lamb)

    # vectorize for performance
    find_minima = np.vectorize(making_top_mask)
    shift_inv_dist_map = find_minima(inv_dist_map)

    # reconstruct
    seed = shift_inv_dist_map
    mask = inv_dist_map
    reconstructed = morph.reconstruction(seed, mask, method="erosion").astype("uint8")

    return reconstructed


def find_maxima(inv_dist_map: np.ndarray, mask: np.ndarray = None) -> np.ndarray:
    """Find all local maxima from 2D image.

    Parameters
    ----------
        inv_dist_map : np.ndarray
            Inverse distance map. Shape (H, W).
        mask : np.ndarray, default=None
            Binary mask to remove small debris. Shape (H, W).

    Returns
    -------
        np.ndarray:
            The found maxima. Shape (H, W).

    """
    reconstructed = h_minima_reconstruction(inv_dist_map, 40)

    res = reconstructed - inv_dist_map
    if mask is not None:
        res[mask == 0] = 0

    return res


def dynamic_ws_alias(
    dist_map: np.ndarray, binary_mask: np.ndarray, thresh: float = 0.5
) -> np.ndarray:
    """Run the dynamic watershed segmentation.

    Minor mods made. (Removed the suspicious stuff from the end.)

    Parameters
    ----------
        dist_map : np.ndarray
            A distance transform. Shape (H, W)
        binary_mask : np.ndarray
            A binary mask. Shape (H, W)
        thresh : float, default=0.5
            The threshold value to find markers from the `dist_map`.

    Returns
    -------
        np.ndarray:
            The labelled instance segmentation result. Shape (H, W).
    """
    # binarize probs and dist map
    binary_dist_map = dist_map > thresh

    # invert distmap
    inv_dist_map = 255 - img_as_ubyte(dist_map)

    # find markers from minima erosion reconstructed maxima of inv dist map
    reconstructed = h_minima_reconstruction(inv_dist_map)
    markers = find_maxima(reconstructed, mask=binary_dist_map)
    markers = ndi.label(markers)[0]

    # apply watershed
    ws = segm.watershed(reconstructed, markers, mask=binary_mask)

    return ws


def post_proc_drfns(
    inst_map: np.ndarray, dist_map: np.ndarray, thresh: float = 0.5, **kwargs
) -> np.ndarray:
    """DRFNS post processing pipeline.

    https://ieeexplore.ieee.org/document/8438559

    Slightly modified. Uses the thresholded prob_map as the mask param
    in watershed. Markers are computed from the regressed distance map
    (inverted).

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        dist_map : np.ndarray
            Distance transform. Shape (H, W)
        thresh : float, default=0.5
            threshold value for markers and binary mask

    Returns
    -------
        np.ndarray:
            The instance labelled segmentation mask. Shape (H, W)
    """
    dist_map = percentile_normalize99(dist_map, amin=0, amax=1)
    binary_mask = binarize(inst_map)
    result = dynamic_ws_alias(dist_map, binary_mask, thresh)

    return result



from typing import Any, Dict, Tuple

import torch

from ...decoders import Decoder
from ...modules.misc_modules import StyleReshape
from ._base_model import BaseMultiTaskSegModel
from ._seg_head import SegHead
from ._timm_encoder import TimmEncoder

__all__ = ["MultiTaskUnet"]


class MultiTaskUnet(BaseMultiTaskSegModel):
    def __init__(
        self,
        decoders: Tuple[str, ...],
        heads: Dict[str, Dict[str, int]],
        n_layers: Dict[str, Tuple[int, ...]],
        n_blocks: Dict[str, Tuple[Tuple[int, ...], ...]],
        out_channels: Dict[str, Tuple[int, ...]],
        long_skips: Dict[str, str],
        dec_params: Dict[str, Tuple[Dict[str, Any], ...]],
        depth: int = 4,
        style_channels: int = 256,
        enc_name: str = "resnet50",
        enc_pretrain: bool = True,
        enc_freeze: bool = False,
        inst_key: str = None,
        aux_key: str = None,
    ) -> None:
        """Create a universal multi-task (2D) unet.

        NOTE: For experimental purposes.

        Parameters
        ----------
            decoders : Tuple[str, ...]
                Names of the decoder branches of this network. E.g. ("cellpose", "sem")
            heads : Dict[str, Dict[str, int]]
                Names of the decoder branches (has to match `decoders`) mapped to dicts
                 of output name - number of output classes. E.g.
                {"cellpose": {"type": 4, "cellpose": 2}, "sem": {"sem": 5}}
            n_layers : Dict[str, Tuple[int, ...]]
                The number of conv layers inside each of the decoder stages.
            n_blocks : Dict[str, Tuple[Tuple[int, ...], ...]]
                The number of blocks inside each conv-layer in each decoder stage.
            out_channels : Tuple[int, ...]
                Out channels for each decoder stage.
            long_skips : Dict[str, str]
                long skip method to be used. One of: "unet", "unetpp", "unet3p",
                "unet3p-lite", None
            dec_params : Dict[str, Tuple[Dict[str, Any], ...]])
                The keyword args for each of the distinct decoder stages. Incudes the
                parameters for the long skip connections and convolutional layers of the
                decoder itself. See the `DecoderStage` documentation for more info.
            depth : int, default=4
                The depth of the encoder. I.e. Number of returned feature maps from
                the encoder. Maximum depth = 5.
            style_channels : int, default=256
                Number of style vector channels. If None, style vectors are ignored.
            enc_name : str, default="resnet50"
                Name of the encoder. See timm docs for more info.
            enc_pretrain : bool, default=True
                Whether to use imagenet pretrained weights in the encoder.
            enc_freeze : bool, default=False
                Freeze encoder weights for training.
            inst_key : str, optional
                The key for the model output that will be used in the instance
                segmentation post-processing pipeline as the binary segmentation result.
            aux_key : str, optional
                The key for the model output that will be used in the instance
                segmentation post-processing pipeline as the auxilliary map.
        """
        super().__init__()
        self.enc_freeze = enc_freeze
        use_style = style_channels is not None
        self.heads = heads
        self.inst_key = inst_key
        self.aux_key = aux_key

        # set timm encoder
        self.encoder = TimmEncoder(
            enc_name,
            depth=depth,
            pretrained=enc_pretrain,
        )

        # style
        self.make_style = None
        if use_style:
            self.make_style = StyleReshape(self.encoder.out_channels[0], style_channels)

            # set decoders
        for decoder_name in decoders:
            decoder = Decoder(
                enc_channels=list(self.encoder.out_channels),
                style_channels=style_channels,
                out_channels=out_channels[decoder_name],
                long_skip=long_skips[decoder_name],
                n_layers=n_layers[decoder_name],
                n_blocks=n_blocks[decoder_name],
                stage_params=dec_params[decoder_name],
            )
            self.add_module(f"{decoder_name}_decoder", decoder)

        # set heads
        for decoder_name in heads.keys():
            for output_name, n_classes in heads[decoder_name].items():
                seg_head = SegHead(
                    in_channels=decoder.out_channels,
                    out_channels=n_classes,
                    kernel_size=1,
                )
                self.add_module(f"{output_name}_seg_head", seg_head)

        self.name = f"MultiTaskUnet-{enc_name}"

        # init decoder weights
        self.initialize()

        # freeze encoder if specified
        if enc_freeze:
            self.freeze_encoder()

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward pass of Multi-task U-net."""
        self._check_input_shape(x)

        feats = self.encoder(x)

        style = None
        if self.make_style is not None:
            style = self.make_style(feats[0])

        dec_feats = self.forward_dec_features(feats, style)

        for decoder_name in self.heads.keys():
            for head_name in self.heads[decoder_name].keys():
                k = self.aux_key if head_name not in dec_feats.keys() else head_name
                dec_feats[head_name] = dec_feats[k]

        out = self.forward_heads(dec_feats)

        return out



from typing import List, Optional

from torch.utils.data import DataLoader, Dataset

from ._basemodule import BaseDataModule


class CustomDataModule(BaseDataModule):
    def __init__(
        self,
        custom_datasets: List[Dataset],
        batch_size: int = 8,
        num_workers: int = 8,
    ) -> None:
        """Set up a custom datamodule with custom datasets.

        Parameters
        ----------
            custom_datasets : List[torch.utils.Dataset]
                A list of initialized torch Datasets. Order: train. valid, test.
            batch_size : int, default=8
                Batch size for the dataloader.
            num_workers : int, default=8
                number of cpu cores/threads used in the dataloading process.
        """
        super().__init__(batch_size, num_workers)
        self.custom_datasets = custom_datasets

    def setup(self, stage: Optional[str] = None) -> None:
        """Set up the train, valid, and test datasets."""
        self.trainset = self.custom_datasets[0]
        self.validset = self.custom_datasets[1]
        self.testset = self.custom_datasets[2]

    def train_dataloader(self) -> DataLoader:
        """Initialize train dataloader."""
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            pin_memory=True,
            num_workers=self.num_workers,
        )

    def val_dataloader(self) -> DataLoader:
        """Initialize val dataloader."""
        return DataLoader(
            self.validset,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=self.num_workers,
        )

    def test_dataloader(self) -> DataLoader:
        """Initialize test dataloader."""
        return DataLoader(
            self.testset,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=self.num_workers,
        )



from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F
from skimage.util import view_as_windows

__all__ = [
    "_get_margins",
    "extract_patches_numpy",
    "stitch_patches_numpy",
    "TilerStitcher",
]


def _get_margins(
    im_height: int, im_width: int, patch_height: int, patch_width: int, pad: int = 200
) -> Tuple[int, int]:
    """Compute margins for the sliding window patcher.

    Parameters
    ----------
        im_heigth : int
            Height of the original image.
        im_width : int
            Width of the original image.
        patch_height : int
            Height of one image patch.
        patch_width : int
            Width of one image patch.
        pad : int, default=200
            Padding in pixels.

    Returns
    -------
        Tuple[int, int]:
            The y- and x-margins in pixels.
    """
    y = np.ceil(((im_height + pad) / patch_height) * patch_height) - im_height
    x = np.ceil(((im_width + pad) / patch_width) * patch_width) - im_width

    return int(y), int(x)


def extract_patches_numpy(
    im: np.ndarray, stride: int, patch_shape: Tuple[int, int, int], padding: bool = True
) -> Tuple[np.ndarray, int, int]:
    """Extract overlapping patches from an image.

    Parameters
    ----------
        im : np.ndarray
            Input image. Shape (H, W, C)|(H, W).
        stride : int
            Stride for the sliding window (pixels).
        patch_shape : Tuple[int, int, int]
            Height, width and num of channels of the patches.
        padding : bool, default=True
            Use reflection padding to make the input shape divisible
            by the stride size.

    Returns
    -------
        Tuple[np.ndarray, int, int]:
            Patched image of shape (N, H, W, C) and the patch grid
            dimensions `ny` and `nx` that are needed for back-stitching.

    Raises
    ------
        ValueError
            If Given number of channels in `patch_shape` does not match
            the number of channels in `im`.
    """
    if len(patch_shape) != 3:
        raise ValueError(
            f"`patch_shape` has to be in (H, W, C) format. Got: {patch_shape}."
        )

    if patch_shape[-1] == 1 and im.ndim == 2:
        im = im[..., None]

    if im.shape[-1] != patch_shape[-1]:
        raise ValueError(
            f"""Mismatch in given input shapes.
            patch_shape[-1] != im.shape[-1].
            patch_shape[-1] = {patch_shape[-1]}.
            im.shape[-1] = {im.shape[-1]}.
            """
        )

    if padding:
        pad_y, pad_x = _get_margins(
            im.shape[0], im.shape[1], patch_shape[0], patch_shape[1]
        )

        im = np.pad(im, [(pad_y, pad_y), (pad_x, pad_x), (0, 0)], mode="reflect")

    patches = view_as_windows(im, patch_shape, stride)  # (ny, nx, 1, H, W, C)

    return (
        patches.reshape(-1, *patch_shape),  # (N, H, W, C)
        patches.shape[0],  # ny
        patches.shape[1],  # nx
    )


def stitch_patches_numpy(
    patches: np.ndarray,
    orig_shape: Tuple[int, ...],
    ny: int,
    nx: int,
    stride: int,
    padding: bool,
) -> np.ndarray:
    """Stitch patches back to original size.

    'Reverse' operation for `extract_patches`.

    Parameters
    ----------
        patches : np.ndarray
            Patched image. Shape (N, pH, pW, C).
        orig_shape : Tuple[int, int, int]
            Shape of the original image. Format (H, W, C).
        ny : int
            Number of rows in patch grid.
        nx : int
            Number of columns in patch grid.
        stride : int
            Stride of the sliding window that was used for patching.
        padding : bool
            Flag, whether padding was used during patching.

    Returns
    -------
        np.ndarray:
            The patches stitched backed to the original size. Shape (H, W, C).

    """
    pad_y = 0
    pad_x = 0
    if padding:
        pad_y, pad_x = _get_margins(orig_shape[0], orig_shape[1], *patches.shape[1:-1])

    pad = stride
    patches = patches.reshape(ny, nx, *patches.shape[1:])
    patches = patches[:, :, 0:pad, 0:pad, :]
    stitched = np.concatenate(np.concatenate(patches, 1), 1)

    stitched = stitched[pad_y : orig_shape[0] + pad_y, pad_x : orig_shape[1] + pad_x, :]

    return stitched


def extract_patches_torch(
    batch: torch.Tensor, stride: int, patch_shape: Tuple[int, int], padding: bool = True
) -> torch.Tensor:
    """Extract patches from a batched tensor.

    Parameters
    ----------
        batch : torch.Tensor
            Batched input images. Shape (B, C, H, W).
        stride : int
            Stride for the sliding window (pixels).
        patch_shape : Tuple[int, int]
            Height, width and num of channels of the patches.
        padding : bool, default=True
            Use reflection padding to make the input shape divisible
            by the stride size.

    Returns
    -------
        torch.Tensor:
            patched batch. Shape (B, C, n_patches, patch_height, patch_width).

    Raises
    ------
        ValueError: If input has wrong shape.
    """
    if batch.dim() != 4:
        raise ValueError(
            f"`batch` shape needs to be in format (B, C, H, W). Got: {batch.shape}."
        )

    if padding:
        pad_y, pad_x = _get_margins(
            batch.shape[-2], batch.shape[-1], patch_shape[0], patch_shape[1]
        )
        batch = F.pad(batch.float(), [pad_x, pad_x, pad_y, pad_y], mode="reflect")

    dims = list(range(2, batch.dim()))
    for dim, patch_size in zip(dims, patch_shape):
        batch = batch.unfold(dim, patch_size, stride)

    patches = batch.contiguous().view(
        batch.shape[0], batch.shape[1], -1, patch_shape[0], patch_shape[1]
    )

    return patches


def stitch_patches_torch(
    patches: torch.Tensor, orig_shape: Tuple[int, int, int], stride: int, padding: bool
) -> torch.Tensor:
    """Stitch tensor patches back to one batched image.

    Parameters
    ----------
        patches : torch.Tensor
            Input tensor patches. Shape (B, C, n_patches, H, W).
        orig_shape : Tuple[int, int, int, int]
            Shape of the original image. Format (B, C, H, W).
        stride : int
            Stride of the sliding window that was used for patching.
        padding : bool
            Flag, whether padding was used during patching.

    Returns
    -------
        torch.Tensor:
            Original sized batch. Shape: (B, C, H, W).
    """
    B, C, _, ph, pw = patches.shape
    H, W = orig_shape[2:]

    pad_y = 0
    pad_x = 0
    if padding:
        pad_y, pad_x = _get_margins(H, W, ph, pw)
        H += pad_y * 2
        W += pad_x * 2

    patches = patches.float().contiguous().view(B, C, -1, ph * pw)
    patches = patches.permute(0, 1, 3, 2)

    patches = patches.contiguous().view(B, C * ph * pw, -1)
    output = F.fold(patches, output_size=(H, W), kernel_size=ph, stride=stride)

    recovery_mask = F.fold(
        torch.ones_like(patches), output_size=(H, W), kernel_size=ph, stride=stride
    )
    output = output / recovery_mask
    output = output[:, :, pad_y : H - pad_y, pad_x : W - pad_x]

    return output


class TilerStitcher:
    def __init__(
        self,
        im_shape: Tuple[int, ...],
        patch_shape: Tuple[int, ...],
        stride: int,
        padding: bool = True,
    ) -> None:
        """Numpy image patcher-stitcher.

        Parameters
        ----------
            im_shape : Tuple[int, ...]
                Input image shape (H, W)|(H, W, C).
            patch_shape : Tuple[int, ...]
                Shape of a patch (pH, pW) | (pH, pW, C).
            stride : int
                Stride size for the sliding window.
            padding : bool, default=True
                Use reflection padding. If set to False, the `backstitch`
                is unlikely able to recover the full sized original image
                but instead cropped image.

        Example
        -------
            >>> im = read_img("/<path>/<to>/img.png)
            >>> print(im.shape)
            (1153, 1307, 3)
            >>> tiler = TilerStitcher(
                    im_shape=im.shape,
                    patch_shape=(256, 256, 3),
                    stride=80,
                    padding=True
                )
            >>> patches = tiler.patch(im)
            >>> print(patches.shape)
            (323, 256, 256, 3)
            >>> orig = tiler.backstitch(patches)
            >>> print(orig.shape)
            (1153, 1307, 3)
        """
        self.im_shape = im_shape
        self.patch_shape = patch_shape
        self.stride = stride
        self.padding = padding
        self.nx = None
        self.ny = None

    def patch(self, im: np.ndarray) -> np.ndarray:
        """Extract patches from an input image.

        Parameters
        ----------
            im : np.ndarray
                Input image. Shape: (H, W)|(H, W, C).

        Returns
        -------
            np.ndarray:
                Patched input image. Shape: (N, pH, pW, C).
        """
        patches, self.ny, self.nx = extract_patches_numpy(
            im, self.stride, self.patch_shape, self.padding
        )

        return patches

    def backstitch(self, patches: np.ndarray) -> np.ndarray:
        """Stitch patches back to the original image size.

        Parameters
        ----------
            patches : np.ndarray
                Patched image. Shape (N, pH, pW, C).

        Returns
        -------
            np.ndarray:
                Backstitched input image. Shape: (H, W, C).
        """
        stitched = stitch_patches_numpy(
            patches, self.im_shape, self.ny, self.nx, self.stride, self.padding
        )
        return stitched.squeeze()


class TilerStitcherTorch:
    def __init__(
        self,
        batch_shape: Tuple[int, int, int, int],
        patch_shape: Tuple[int, int],
        stride: int,
        padding: bool = True,
    ) -> None:
        """Patch extractor and stitcher class.

        Operates on batched torch tensors.
        I.e tensors of shape: (B, C, H, W).

        Parameters
        ----------
            batch_shape : Tuple[int, int, int, int]
                Input batch image shape. Has to be in format: (B, C, H, W).
            patch_shape : Tuple[int, int]
                Shape of one patch (pH, pW).
            stride : int
                Stride size for the sliding window.
            padding : bool, default=True
                Use reflection padding (To get a square image.). If set to False,
                the `backstitch` is unlikely able to recover the full sized original
                image but instead cropped image.

        Example
        -------
            >>> t = torch.rand([2, 1, 1100, 1200])
            >>> ts = TilerStitcherTorch((2, 1, 1100, 1200), (256, 256), 80, padding=1)
            >>> patches = ts.patch(t)
            >>> print(patches.shape)
            torch.Size([2, 1, 272, 256, 256])

            >>> stitched = ts.backstitch(patches)
            >>> print(stitched.shape)
            torch.Size([2, 1, 1100, 1200])
        """
        self.batch_shape = batch_shape
        self.patch_shape = patch_shape
        self.stride = stride
        self.padding = padding

    def patch(self, im: torch.Tensor) -> torch.Tensor:
        """Extract patches from an input image.

        Parameters
        ----------
            im : torch.Tensor
                Input image. Shape: (B, C, H, W).

        Returns
        -------
            torch.Tensor:
                Patched input. Shape: (B, C, n_patches, patch_height, patch_width).
        """
        patches = extract_patches_torch(im, self.stride, self.patch_shape, self.padding)

        return patches

    def backstitch(self, patches: torch.Tensor) -> torch.Tensor:
        """Stitch patches back to the original image size.

        Parameters
        ----------
            patches : torch.Tensor
                Patched image. Shape: (B, C, n_patches, patch_height, patch_width).

        Returns
        -------
            np.ndarray:
                Backstitched input image. Shape: (B, C, H, W).
        """
        stitched = stitch_patches_torch(
            patches, self.batch_shape, self.stride, self.padding
        )

        return stitched



import numpy as np
from skimage import img_as_ubyte

__all__ = [
    "percentile_normalize",
    "percentile_normalize99",
    "normalize",
    "minmax_normalize",
    "float2ubyte",
]


def percentile_normalize(
    img: np.ndarray, lower: float = 0.01, upper: float = 99.99
) -> np.ndarray:
    """Channelwise percentile normalization to range [0, 1].

    Parameters
    ----------
        img : np.ndarray:
            Input image to be normalized. Shape (H, W, C)|(H, W).
        lower : float, default=0.01:
            The lower percentile
        upper : float, default=99.99:
            The upper percentile

    Returns
    -------
        np.ndarray:
            Normalized img. Same shape as input. dtype: float32.

    Raises
    ------
        ValueError
            If input image does not have shape (H, W) or (H, W, C).
    """
    axis = (0, 1)

    if img.ndim not in (2, 3):
        raise ValueError(
            f"Input img needs to have shape (H, W, C)|(H, W). Got: {img.shape}"
        )

    im = img.copy()

    upercentile = np.percentile(im, upper)
    lpercentile = np.percentile(im, lower)

    return np.interp(im, (lpercentile, upercentile), axis).astype(np.float32)


def percentile_normalize99(
    img: np.ndarray, amin: float = None, amax: float = None
) -> np.ndarray:
    """Channelwise 1-99 percentile normalization. Optional clamping.

    Parameters
    ----------
        img : np.ndarray:
            Input image to be normalized. Shape (H, W, C)|(H, W).
        amin : float, optional
            Clamp min value. No clamping performed if None.
        amax : float, optional
            Clamp max value. No clamping performed if None.

    Returns
    -------
        np.ndarray:
            Normalized image. Same shape as input. dtype: float32.

    Raises
    ------
        ValueError
            If input image does not have shape (H, W) or (H, W, C).
    """
    axis = (0, 1)

    if img.ndim not in (2, 3):
        raise ValueError(
            f"Input img needs to have shape (H, W, C)|(H, W). Got: {img.shape}"
        )

    im = img.copy()
    percentile1 = np.percentile(im, q=1, axis=axis)
    percentile99 = np.percentile(im, q=99, axis=axis)
    im = (im - percentile1) / (percentile99 - percentile1 + 1e-7)

    # clamp
    if not any(x is None for x in (amin, amax)):
        im = np.clip(im, a_min=amin, a_max=amax)

    return im.astype(np.float32)


def normalize(
    img: np.ndarray, standardize: bool = True, amin: float = None, amax: float = None
) -> np.ndarray:
    """Channelwise mean centering or standardizing of an image. Optional clamping.

    Parameters
    ----------
        img : np.ndarray
            Input image to be normalized. Shape (H, W, C)|(H, W).
        standardize: bool, default=True
            If True, divide with standard deviation after mean centering
        amin : float, optional
            Clamp min value. No clamping performed if None.
        amax : float, optional
            Clamp max value. No clamping performed if None.

    Returns
    -------
        np.ndarray:
            Normalized image. Same shape as input. dtype: float32.

    Raises
    ------
        ValueError
            If input image does not have shape (H, W) or (H, W, C).
    """
    axis = (0, 1)

    if img.ndim not in (2, 3):
        raise ValueError(
            f"Input img needs to have shape (H, W, C)|(H, W). Got: {img.shape}"
        )

    im = img.copy()

    # mean center
    im = im - im.mean(axis=axis, keepdims=True)

    if standardize:
        im /= im.std(axis=axis, keepdims=True)

    # clamp
    if not any(x is None for x in (amin, amax)):
        im = np.clip(im, a_min=amin, a_max=amax)

    return im.astype(np.float32)


def minmax_normalize(
    img: np.ndarray, amin: float = None, amax: float = None
) -> np.ndarray:
    """Min-max normalization per image channel. Optional clamping.

    Parameters
    ----------
        img : np.ndarray:
            Input image to be normalized. Shape (H, W, C)|(H, W).
        amin : float, optional
            Clamp min value. No clamping performed if None.
        amax : float, optional
            Clamp max value. No clamping performed if None.

    Returns
    -------
        np.ndarray:
            Min-max normalized image. Same shape as input. dtype: float32.

    Raises
    ------
        ValueError
            If input image does not have shape (H, W) or (H, W, C).
    """
    if img.ndim not in (2, 3):
        raise ValueError(
            f"Input img needs to have shape (H, W, C)|(H, W). Got: {img.shape}"
        )

    im = img.copy()
    im = (im - im.min()) / (im.max() - im.min())

    # clamp
    if not any(x is None for x in (amin, amax)):
        im = np.clip(im, a_min=amin, a_max=amax)

    return im.astype(np.float32)


def float2ubyte(mat: np.ndarray, normalize: bool = False) -> np.ndarray:
    """Convert float64 to uint8.

    Float matrix values need to be in range [-1, 1] for img_as_ubyte so
    the image is normalized or clamped before conversion.

    Parameters
    ----------
        mat : np.ndarray
            A float64 matrix. Shape (H, W, C).
        normalize (bool, default=False):
            Normalizes input to [0, 1] first. If not True,
            clips values between [-1, 1].

    Returns
    -------
        np.ndarray:
            A uint8 matrix. Shape (H, W, C). dtype: uint8.
    """
    m = mat.copy()

    if normalize:
        m = minmax_normalize(m)
    else:
        m = np.clip(m, a_min=-1, a_max=1)

    return img_as_ubyte(m)



import numpy as np
import pytest

from cellseg_models_pytorch.transforms.albu_transforms import (
    apply_each,
    binarize_transform,
    cellpose_transform,
    compose,
    contour_transform,
    dist_transform,
    edgeweight_transform,
    hovernet_transform,
    omnipose_transform,
    smooth_dist_transform,
    stardist_opt_transform,
    stardist_transform,
)


@pytest.mark.parametrize(
    "transform",
    [cellpose_transform, hovernet_transform, omnipose_transform],
)
@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14), np.int32)])
def test_gradient_transforms(inst_map, transform, zero_input):
    trans = compose([transform()])

    if zero_input is not None:
        inst_map = zero_input

    transformed = trans(image=inst_map, inst_map=inst_map)

    assert transformed["inst_map"].dtype == np.float64
    assert transformed["inst_map"].shape == (2,) + inst_map.shape


@pytest.mark.parametrize(
    "transform",
    [dist_transform, smooth_dist_transform, contour_transform, edgeweight_transform],
)
@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14))])
def test_inst_transforms(inst_map, transform, zero_input):
    trans = compose([transform()])

    if zero_input is not None:
        inst_map = zero_input

    transformed = trans(image=inst_map, inst_map=inst_map)

    assert transformed["inst_map"].dtype == np.float64
    assert transformed["inst_map"].shape == inst_map.shape


@pytest.mark.parametrize("transform", [binarize_transform])
@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14))])
def test_binarize_transforms(inst_map, transform, zero_input):
    trans = compose([transform()])

    if zero_input is not None:
        inst_map = zero_input

    transformed = trans(image=inst_map, inst_map=inst_map)

    assert transformed["inst_map"].dtype == np.uint8
    assert transformed["inst_map"].shape == inst_map.shape


@pytest.mark.parametrize("transform", [stardist_transform, stardist_opt_transform])
@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14), np.int32)])
def test_stardist_transforms(inst_map, transform, zero_input):
    trans = compose([transform(n_rays=3, n_segments=10)])

    if zero_input is not None:
        inst_map = zero_input

    transformed = trans(image=inst_map, inst_map=inst_map)

    assert transformed["inst_map"].dtype == np.float32
    assert transformed["inst_map"].shape == (3,) + inst_map.shape


@pytest.mark.parametrize("zero_input", [None, np.zeros((14, 14))])
def test_inst_transform_pipeline(inst_map, zero_input):
    pipeline = apply_each(
        [edgeweight_transform(), cellpose_transform(), smooth_dist_transform()]
    )

    if zero_input is not None:
        inst_map = zero_input

    transformed = pipeline(image=inst_map, inst_map=inst_map)

    assert transformed["edgeweight"]["inst_map"].dtype == np.float64
    assert transformed["cellpose"]["inst_map"].dtype == np.float64
    assert transformed["smoothdist"]["inst_map"].dtype == np.float64



from typing import Any, Dict, Optional

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F

try:
    import wandb
except ImportError:
    raise ImportError("wandb required. `pip install wandb`")

from ..functional import iou

__all__ = ["WandbImageCallback", "WandbClassMetricCallback"]


class WandbImageCallback(pl.Callback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        freq: int = 100,
    ) -> None:
        """Create a callback that logs prediction masks to wandb."""
        super().__init__()
        self.freq = freq
        self.type_classes = type_classes
        self.sem_classes = sem_classes

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        if batch_idx % self.freq == 0:
            outputs = outputs["soft_masks"]

            log_dict = {
                "global_step": trainer.global_step,
                "epoch": trainer.current_epoch,
            }

            img = batch["image"].detach().to("cpu").numpy()

            if "type" in list(batch.keys()):
                type_target = batch["type"].detach().to("cpu").numpy()
                soft_types = outputs["type"].detach().to("cpu")
                types = torch.argmax(F.softmax(soft_types, dim=1), dim=1).numpy()

                log_dict["val/cell_types"] = [
                    wandb.Image(
                        im.transpose(1, 2, 0),
                        masks={
                            "predictions": {
                                "mask_data": t,
                                "class_labels": self.type_classes,
                            },
                            "ground_truth": {
                                "mask_data": tt,
                                "class_labels": self.type_classes,
                            },
                        },
                    )
                    for im, t, tt in zip(img, types, type_target)
                ]

            if "sem" in list(batch.keys()):
                sem_target = batch["sem"].detach().to("cpu").numpy()
                soft_sem = outputs["sem"].detach().to(device="cpu")
                sem = torch.argmax(F.softmax(soft_sem, dim=1), dim=1).numpy()

                log_dict["val/tissue_areas"] = [
                    wandb.Image(
                        im.transpose(1, 2, 0),
                        masks={
                            "predictions": {
                                "mask_data": s,
                                "class_labels": self.sem_classes,
                            },
                            "ground_truth": {
                                "mask_data": st,
                                "class_labels": self.sem_classes,
                            },
                        },
                    )
                    for im, s, st in zip(img, sem, sem_target)
                ]

            for m in list(batch.keys()):
                if m not in ("sem", "type", "inst", "image"):
                    aux = outputs[m].detach().to(device="cpu")
                    log_dict[f"val/{m}"] = [
                        wandb.Image(a[i, ...], caption=f"{m} maps")
                        for a in aux
                        for i in range(a.shape[0])
                    ]

            trainer.logger.experiment.log(log_dict)


class WandbClassMetricCallback(pl.Callback):
    def __init__(
        self,
        type_classes: Dict[str, int],
        sem_classes: Optional[Dict[str, int]],
        freq: int = 100,
        return_series: bool = True,
        return_bar: bool = True,
        return_table: bool = False,
    ) -> None:
        """Call back to compute per-class ious and log them to wandb."""
        super().__init__()
        self.type_classes = type_classes
        self.sem_classes = sem_classes
        self.freq = freq
        self.return_series = return_series
        self.return_bar = return_bar
        self.return_table = return_table
        self.cell_ious = np.empty(0)
        self.sem_ious = np.empty(0)

    def compute(
        self,
        key: str,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
    ) -> np.ndarray:
        """Compute the iou per class."""
        target = batch[key].detach()
        soft_types = outputs[key].detach()
        pred = F.softmax(soft_types, dim=1)

        met = iou(pred, target).mean(dim=0)
        return met.to("cpu").numpy()

    def get_table(
        self, ious: np.ndarray, x: np.ndarray, classes: Dict[int, str]
    ) -> wandb.Table:
        """Return a wandb Table with step, iou and label values for every step."""
        batch_data = [
            [xi * self.freq, c, np.round(ious[xi, i], 4)]
            for i, c, in classes.items()
            for xi in x
        ]

        return wandb.Table(data=batch_data, columns=["step", "label", "value"])

    def get_bar(self, iou: np.ndarray, classes: Dict[int, str], title: str) -> Any:
        """Return a wandb bar plot object of the current per class iou values."""
        batch_data = [[lab, val] for lab, val in zip(list(classes.values()), iou)]
        table = wandb.Table(data=batch_data, columns=["label", "value"])
        return wandb.plot.bar(table, "label", "value", title=title)

    def get_series(
        self, ious: np.ndarray, x: np.ndarray, classes: Dict[int, str], title: str
    ) -> Any:
        """Return a wandb series plot obj of the per class iou values over timesteps."""
        return wandb.plot.line_series(
            xs=x.tolist(),
            ys=[ious[:, c].tolist() for c in classes.keys()],
            keys=list(classes.values()),
            title=title,
            xname="step",
        )

    def batch_end(
        self,
        trainer: pl.Trainer,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        phase: str,
    ) -> None:
        """Log metrics at every 100th step to wandb."""
        if batch_idx % self.freq == 0:
            log_dict = {}
            if "type" in list(batch.keys()):
                iou = self.compute("type", outputs, batch)
                self.cell_ious = np.append(self.cell_ious, iou)
                cell_ious = self.cell_ious.reshape(-1, len(self.type_classes))
                x = np.arange(cell_ious.shape[0])

                if self.return_table:
                    log_dict[f"{phase}/type_ious_table"] = self.get_table(
                        cell_ious, x, self.type_classes
                    )

                if self.return_series:
                    log_dict[f"{phase}/type_ious_per_class"] = self.get_series(
                        cell_ious, x, self.type_classes, title="Per type class mIoU"
                    )

                if self.return_bar:
                    log_dict[f"{phase}/type_ious_bar"] = self.get_bar(
                        list(iou), self.type_classes, title="Cell class mIoUs"
                    )

            if "sem" in list(batch.keys()):
                iou = self.compute("sem", outputs, batch)

                self.sem_ious = np.append(self.sem_ious, iou)
                sem_ious = self.sem_ious.reshape(-1, len(self.sem_classes))
                x = np.arange(sem_ious.shape[0])

                if self.return_table:
                    log_dict[f"{phase}/sem_ious_table"] = self.get_table(
                        cell_ious, x, self.type_classes
                    )

                if self.return_series:
                    log_dict[f"{phase}/sem_ious_per_class"] = self.get_series(
                        cell_ious, x, self.type_classes, title="Per sem class mIoU"
                    )

                if self.return_bar:
                    log_dict[f"{phase}/sem_ious_bar"] = self.get_bar(
                        list(iou), self.type_classes, title="Sem class mIoUs"
                    )

            trainer.logger.experiment.log(log_dict)

    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        self.batch_end(trainer, outputs["soft_masks"], batch, batch_idx, phase="train")

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Log the inputs and outputs of the model to wandb."""
        self.batch_end(trainer, outputs["soft_masks"], batch, batch_idx, phase="val")



import numpy as np
import pytest

from cellseg_models_pytorch.metrics import (
    accuracy,
    dice_coef,
    f1score,
    get_stats,
    iou_score,
)


@pytest.fixture(scope="package")
def binary_true() -> np.ndarray:
    true = true = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 0, 0, 1, 1, 1],
            [1, 1, 0, 0, 0, 0, 1, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 0, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 1, 0, 0],
        ],
        dtype=int,
    )

    return true


@pytest.fixture(scope="package")
def binary_pred() -> np.ndarray:
    pred = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 0, 0, 1, 1, 1],
            [1, 1, 0, 1, 1, 0, 1, 1],
            [1, 0, 0, 1, 1, 0, 0, 1],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 0, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
        ],
        dtype=int,
    )
    return pred


def test_get_stats(binary_true, binary_pred):
    tp, fp, fn = get_stats(binary_true, binary_pred)

    assert tp.sum() == 32
    assert fp.sum() == 4
    assert fn.sum() == 2


def test_iou(binary_true, binary_pred):
    tp, fp, fn = get_stats(binary_true, binary_pred)
    iou = iou_score(tp, fp, fn)

    expected = 0.84
    np.testing.assert_allclose(iou, expected, rtol=1e-2, err_msg=f"{iou} != {expected}")


def test_accuracy(binary_true, binary_pred):
    tp, fp, fn = get_stats(binary_true, binary_pred)
    acc = accuracy(tp, fp, fn)

    expected = 0.90
    np.testing.assert_allclose(acc, expected, rtol=1e-2, err_msg=f"{acc} != {expected}")


def test_dice(binary_true, binary_pred):
    tp, fp, fn = get_stats(binary_true, binary_pred)
    dice = dice_coef(tp, fp, fn)

    expected = 0.91
    np.testing.assert_allclose(
        dice, expected, rtol=1e-2, err_msg=f"{dice} != {expected}"
    )


def test_f1score(binary_true, binary_pred):
    tp, fp, fn = get_stats(binary_true, binary_pred)
    f1 = f1score(tp, fp, fn)

    expected = 0.91
    np.testing.assert_allclose(f1, expected, rtol=1e-2, err_msg=f"{f1} != {expected}")



from itertools import chain
from typing import Any, Dict, List, Tuple

import torch
import torch.nn as nn

from ._initialization import initialize_decoder, initialize_head

__all__ = ["BaseMultiTaskSegModel"]


class BaseMultiTaskSegModel(nn.ModuleDict):
    def forward_dec_features(
        self, feats: List[torch.Tensor], style: torch.Tensor = None
    ) -> Dict[str, torch.Tensor]:
        """Forward pass of the decoders in a multi-task seg model."""
        res = {}
        decoders = [k for k in self.keys() if "decoder" in k]

        for dec in decoders:
            x = self[dec](*feats, style=style)
            branch = dec.split("_")[0]
            res[branch] = x

        return res

    def forward_heads(
        self, dec_feats: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Forward pass of the seg heads in a multi-task seg model."""
        res = {}
        heads = [k for k in self.keys() if "head" in k]

        for head in heads:
            branch = head.split("_")[0]
            x = self[head](dec_feats[branch])
            res[branch] = x

        return res

    def initialize(self) -> None:
        """Init the decoder branches and their classification/regression heads."""
        for name, module in self.items():
            if "decoder" in name:
                initialize_decoder(module)
            if "head" in name:
                initialize_head(module)

    def freeze_encoder(self) -> None:
        """Freeze the parameters of the encoeder."""
        for param in self.encoder.parameters():
            param.requires_grad = False

    def _check_input_shape(self, x: torch.Tensor) -> None:
        """Check that the input is divisible by 32."""
        h, w = x.shape[-2:]

        if (h % 32) + (w % 32):
            raise RuntimeError(
                "Illegal input shape. Expected input H/W to be divisible by 32. "
                f"Got height: {h}, and width: {w}."
            )

    def _check_decoder_args(
        self, decoders: Tuple[str], must_haves: Tuple[str, ...]
    ) -> str:
        """Check that the decoders arg contains needed values."""
        if not any([d in must_haves for d in decoders]):
            raise ValueError(
                f"`decoders` need to contain one of: {must_haves} " f"Got: {decoders}."
            )

        if len(must_haves) > 1:
            if all([m in decoders for m in must_haves]):
                raise ValueError(
                    f"`decoders` need to contain only one of: {must_haves} "
                    f"Got: {decoders}."
                )

        for must_have in must_haves:
            try:
                ix = decoders.index(must_have)
                return decoders[ix]
            except ValueError:
                pass

    def _get_inner_keys(self, d: Dict[str, Dict[str, Any]]) -> List[str]:
        """Get the inner dict keys from a nested dict."""
        return list(chain.from_iterable(list(d[k].keys()) for k in d.keys()))

    def _flatten_inner_dicts(self, d: Dict[str, Dict[str, Any]]) -> List[str]:
        """Get the inner dicts as one dict from a nested dict."""
        return dict(chain.from_iterable(list(d[k].items()) for k in d.keys()))

    def _check_head_args(
        self, heads: Dict[str, int], decoders: Tuple[str, ...]
    ) -> None:
        """Check for illegal `heads` args."""
        if not set(decoders) == set(heads.keys()):
            raise ValueError(
                "The decoder names need match exactly to the keys of `heads`. "
                f"Got decoders: {decoders} and heads: {list(heads.keys())}."
            )

        for head in heads.keys():
            if not any([h == head for h in heads[head].keys()]):
                raise ValueError(
                    "For every decoder name one matching head name has to exist "
                    f"Got heads: {[j for h, i in heads.items() for j in i.keys()]}"
                    f"decoders: {decoders}."
                )

    def _check_depth(self, depth: int, arrs: Dict[str, Tuple[Any, ...]]) -> None:
        """Check that the depth matches to tuple args."""
        if not 3 <= depth <= 5:
            raise ValueError(
                f"max value for `depth` is 5, min value is 3. Got: {depth}"
            )

        for name, arr in arrs.items():
            if depth != len(arr):
                raise ValueError(
                    f"The length of `{name}` should be equal to arg `depth`: {depth}. "
                    f"For `{name}`, got: {arr}."
                )



import pytest
import torch

from cellseg_models_pytorch.training.functional.train_metrics import accuracy, iou


@pytest.mark.parametrize("metric", [accuracy, iou])
@pytest.mark.parametrize("map_type", ["type", "sem"])
@pytest.mark.parametrize("activation", ["sigmoid", "softmax"])
def test_metrics(type_map_tensor, sem_map_tensor, metric, map_type, activation):
    if map_type == "type":
        target = type_map_tensor
        n_channels = 7
    else:
        target = sem_map_tensor
        n_channels = 5

    yhat = torch.rand(8, n_channels, 320, 320)
    metric(yhat, target, activation)



import warnings

from torch.optim import ASGD, SGD, Adadelta, Adagrad, Adam, Adamax, AdamW, RMSprop
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    CyclicLR,
    ExponentialLR,
    LambdaLR,
    ReduceLROnPlateau,
)

from .utils import adjust_optim_params

EXTRA_OPTIM_LOOKUP = {}
try:
    from torch_optimizer import (
        PID,
        QHM,
        SGDW,
        AccSGD,
        AdaBelief,
        AdaBound,
        AdaMod,
        AdamP,
        Apollo,
        DiffGrad,
        Lamb,
        Lookahead,
        NovoGrad,
        QHAdam,
        RAdam,
        Ranger,
        RangerQH,
        RangerVA,
        Yogi,
    )

    EXTRA_OPTIM_LOOKUP = {
        "accsgd": AccSGD,
        "adabound": AdaBound,
        "adabelief": AdaBelief,
        "adamp": AdamP,
        "apollo": Apollo,
        "adamod": AdaMod,
        "diffgrad": DiffGrad,
        "lamb": Lamb,
        "novograd": NovoGrad,
        "pid": PID,
        "qhadam": QHAdam,
        "qhm": QHM,
        "radam": RAdam,
        "sgwd": SGDW,
        "yogi": Yogi,
        "ranger": Ranger,
        "rangerqh": RangerQH,
        "rangerva": RangerVA,
        "lookahead": Lookahead,
    }
except ModuleNotFoundError:
    warnings.warn(
        "`torch_optimizer` optimzers not available. To use them, install with "
        "`pip install torch-optimizer`."
    )

SCHED_LOOKUP = {
    "lambda": LambdaLR,
    "reduce_on_plateau": ReduceLROnPlateau,
    "cyclic": CyclicLR,
    "exponential": ExponentialLR,
    "cosine_annealing": CosineAnnealingLR,
    "cosine_annealing_warm": CosineAnnealingWarmRestarts,
}

OPTIM_LOOKUP = {
    "adam": Adam,
    "rmsprop": RMSprop,
    "sgd": SGD,
    "adadelta": Adadelta,
    "adagrad": Adagrad,
    "adamax": Adamax,
    "adamw": AdamW,
    "asgd": ASGD,
    **EXTRA_OPTIM_LOOKUP,
}

__all__ = [
    "SCHED_LOOKUP",
    "OPTIM_LOOKUP",
    "LambdaLR",
    "ReduceLROnPlateau",
    "CosineAnnealingLR",
    "CosineAnnealingWarmRestarts",
    "CyclicLR",
    "ExponentialLR",
    "Adam",
    "RMSprop",
    "SGD",
    "Adadelta",
    "Adagrad",
    "Adamax",
    "AdamW",
    "ASGD",
    "AccSGD",
    "AdaBound",
    "AdaBelief",
    "AdamP",
    "Apollo",
    "AdaMod",
    "DiffGrad",
    "Lamb",
    "NovoGrad",
    "PID",
    "QHAdam",
    "QHM",
    "RAdam",
    "SGDW",
    "Yogi",
    "Ranger",
    "RangerQH",
    "RangerVA",
    "Lookahead",
    "adjust_optim_params",
]



from typing import Any, Dict, Tuple

__all__ = ["_create_cellpose_args"]


def _create_cellpose_args(
    layer_depth: Tuple[int, ...],
    norm: str,
    act: str,
    conv: str,
    att: str,
    preact: bool,
    preattend: bool,
    short_skip: str,
    use_style: bool,
    merge_policy: str,
    skip_params: Dict[str, Any],
) -> Tuple[Dict[str, Any], ...]:
    """Create the args to build CellPose-Unet architecture."""
    skip_params = skip_params if skip_params is not None else {"k": None}

    return tuple(
        {
            "layer_residual": True,
            "merge_policy": merge_policy,
            "short_skips": (short_skip,),
            "block_types": (("basic",) * ld,),
            "kernel_sizes": ((3,) * ld,),
            "expand_ratios": ((1.0,) * ld,),
            "groups": ((1,) * ld,),
            "biases": ((False,) * ld,),
            "normalizations": ((norm,) * ld,),
            "activations": ((act,) * ld,),
            "convolutions": ((conv,) * ld,),
            "attentions": ((att,) + (None,) * (ld - 1),),
            "preactivates": ((preact,) * ld,),
            "preattends": ((preattend,) * ld,),
            "use_styles": ((use_style,) * (ld - 1) + (False,),),
            "skip_params": {
                "short_skips": (short_skip,),
                "block_types": (("basic",),),
                "convolutions": ((conv,),),
                "normalizations": ((norm,),),
                "activations": ((act,),),
                **skip_params,
            },
        }
        for ld in layer_depth
    )



from ._base_inferer import BaseInferer
from .folder_dataset import FolderDataset
from .post_processor import PostProcessor
from .predictor import Predictor
from .resize_inferer import ResizeInferer
from .sliding_window_inferer import SlidingWindowInferer

__all__ = [
    "BaseInferer",
    "Predictor",
    "PostProcessor",
    "ResizeInferer",
    "SlidingWindowInferer",
    "FolderDataset",
]



from typing import Callable, Dict, List, Tuple

import albumentations as A
import numpy as np
import torch
from albumentations.core.transforms_interface import BasicTransform
from albumentations.pytorch import ToTensorV2

__all__ = ["apply_each", "compose", "to_tensor", "to_tensorv3"]


class OnlyInstMapTransform(A.BasicTransform):
    """Transforms applied to only instance labelled masks."""

    def __init__(self) -> None:
        super().__init__(always_apply=True, p=1.0)

    @property
    def targets(self) -> Dict[str, Callable]:
        return {
            "image": self.apply,  # needs to have this key or braks.
            "inst_map": self.apply_to_instmap,
        }

    def apply(self, img: np.ndarray, **params):
        return

    def apply_to_instmap(self, inst_map: np.ndarray, **params):
        raise NotImplementedError("`apply_to_instmap` method not implemented")


class ApplyEach(A.BaseCompose):
    def __init__(self, transforms: List[A.BasicTransform], p: float = 1.0) -> None:
        """Apply each transform to the input non-sequentially.

        Returns outputs for each transform.
        """
        super().__init__(transforms, p)

    def __call__(self, **data):
        res = {}
        for t in self.transforms:
            res[t.name] = t(force_apply=True, **data)

        return res


class ToTensorV3(A.BasicTransform):
    """Convert image, masks and auxilliary inputs to tensor."""

    def __init__(self):
        super().__init__(always_apply=True, p=1.0)

    @property
    def targets(self) -> Dict[str, Callable]:
        return {
            "image": self.apply,
            "masks": self.apply_to_masks,
            "aux": self.apply_to_aux,
        }

    def apply(self, img: np.ndarray, **params):
        if len(img.shape) not in [2, 3]:
            raise ValueError("Images need to be in HW or HWC format")

        if len(img.shape) == 2:
            img = np.expand_dims(img, 2)

        return torch.from_numpy(img.transpose(2, 0, 1)).float()

    def apply_to_masks(self, masks: List[np.ndarray], **params):
        for i, mask in enumerate(masks):
            if mask.ndim == 3:
                mask = mask.transpose(2, 0, 1)
            masks[i] = torch.from_numpy(mask).long()

        return masks

    def apply_to_aux(self, auxilliary: Dict[str, np.ndarray], **params):
        for k, aux in auxilliary.items():
            if k != "binary":
                auxilliary[k] = torch.from_numpy(aux["inst_map"]).float()
            else:
                auxilliary[k] = torch.from_numpy(aux["inst_map"]).long()

        return auxilliary

    def get_transform_init_args_names(self) -> Tuple[str, ...]:
        return ("always_apply", "p")


def apply_each(transforms: List[OnlyInstMapTransform]) -> Callable:
    """Apply each transform wrapper.

    Example
    -------
        >>> im = read_image("/path/to/image.png")
        >>> inst_map = read_mask("/path/to/mask.mat")
        >>> tr = apply_each([cellpose_transform(), edgeweight_transform()])
        >>> aug = tr(image=im, inst_map=inst_map)
        >>> print(aug["cellpose"]["inst_map"].shape)
        (2, 256, 256)
        >>> print(aug["edgeweight"]["inst_map"].shape)
        (256, 256)

    Returns
    -------
        ApplyEach: ApplyEach object.
    """
    result = ApplyEach([item for sublist in transforms for item in sublist])

    return result


def compose(transforms_to_compose: List[A.BasicTransform]) -> Callable:
    """Compose transforms with albumentations Compose.

    Takes in a list of albumentation transforms and composes them to one
    transformation pipeline.

    Example
    -------
        >>> im = read_image("/path/to/image.png")
        >>> inst_map = read_mask("/path/to/mask.mat")
        >>> tr = compose([rigid_transform(), blur_transform(), minmax_normalize()])
        >>> aug = tr(image=im, masks=[inst_map])
        >>> print(aug["image"].shape)
        (256, 256, 3)
        >>> print(aug["masks"][0].shape)
        (256, 256)

    Returns
    -------
        Callable:
            A composed pipeline of albumentation transforms.
    """
    result = A.Compose([item for sublist in transforms_to_compose for item in sublist])
    return result


def to_tensor(**kwargs) -> List[BasicTransform]:
    """Convert each patch (np.ndarray) is into torch.Tensor.

    Returns
    -------
        List[BasicTransform]:
            A tensor conversion transform.
    """
    return [ToTensorV2()]


def to_tensorv3(**kwargs) -> BasicTransform:
    """Convert images, label-, auxilliary-, and semantic masks to tensors.

    Returns
    -------
        BasicTransform:
            A tensor conversion transform
    """
    return ToTensorV3()



import zipfile
from pathlib import Path
from typing import Union

try:
    import pytorch_lightning as pl
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the `csmp.datamodules` module, the pytorch-lightning lib, is needed. "
        "Install with `pip install pytorch-lightning`"
    )
from torch.utils.data import DataLoader

__all__ = ["BaseDataModule"]


class BaseDataModule(pl.LightningDataModule):
    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 8,
    ) -> None:
        """Set up pannuke datamodule..

        Parameters
        ----------
            batch_size : int, default=8
                Batch size for the dataloader.
            num_workers : int, default=8
                number of cpu cores/threads used in the dataloading process.
        """
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers

    def train_dataloader(self) -> DataLoader:
        """Initialize train dataloader."""
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            pin_memory=True,
            num_workers=self.num_workers,
        )

    def val_dataloader(self) -> DataLoader:
        """Initialize val dataloader."""
        return DataLoader(
            self.validset,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=self.num_workers,
        )

    def test_dataloader(self) -> DataLoader:
        """Initialize test dataloader."""
        return DataLoader(
            self.testset,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=self.num_workers,
        )

    @staticmethod
    def extract_zips(path: Union[str, Path], rm: bool = False) -> None:
        """Extract files from all the .zip files inside a folder.

        Parameters
        ----------
            path : str or Path
                Path to a folder containing .zip files.
            rm :bool, default=False
                remove the .zip files after extraction.
        """
        for f in Path(path).iterdir():
            if f.is_file() and f.suffix == ".zip":
                with zipfile.ZipFile(f, "r") as z:
                    z.extractall(path)
                if rm:
                    f.unlink()



"""
Adapted from HoVer-Net repo. Code style mods.

https://github.com/vqdang/hover_net/blob/master/models/hovernet/post_proc.py

MIT License

Copyright (c) 2020 vqdang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import cv2
import numpy as np
import scipy.ndimage as ndi
import skimage.segmentation as segm

from cellseg_models_pytorch.utils import (
    percentile_normalize99,
    remove_debris_instance,
    remove_small_objects,
)

__all__ = ["post_proc_hovernet"]


def post_proc_hovernet(
    inst_map: np.ndarray, hover_map: np.ndarray, enhance: bool = True, **kwargs
) -> np.ndarray:
    """HoVer-Net post processing pipeline.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        hover_map : np.ndarray
            Regressed horizontal and vertical gradients. Shape: (2, H, W).
        enhance : bool, default=True
            Normalizes hover-maps to the 0-99 percentiles and clamps the
            values to min=-1 and max=1.

    Returns
    -------
        np.ndarray:
            Post-processed inst map. Shape (H, W). Dtype: int32
    """
    v_dir = cv2.normalize(
        hover_map[0],
        None,
        alpha=0,
        beta=1,
        norm_type=cv2.NORM_MINMAX,
        dtype=cv2.CV_32F,
    )
    h_dir = cv2.normalize(
        hover_map[1],
        None,
        alpha=0,
        beta=1,
        norm_type=cv2.NORM_MINMAX,
        dtype=cv2.CV_32F,
    )

    if enhance:
        h_dir = percentile_normalize99(h_dir, amin=-1, amax=1)
        v_dir = percentile_normalize99(v_dir, amin=-1, amax=1)

    sobelh = cv2.Sobel(h_dir, cv2.CV_64F, 1, 0, ksize=21)
    sobelv = cv2.Sobel(v_dir, cv2.CV_64F, 0, 1, ksize=21)

    sobelh = 1 - (
        cv2.normalize(
            sobelh, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
    )
    sobelv = 1 - (
        cv2.normalize(
            sobelv, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
    )

    overall = np.maximum(sobelh, sobelv)
    overall = overall - (1 - inst_map)
    overall[overall < 0] = 0

    dist = (1.0 - overall) * inst_map
    dist = -cv2.GaussianBlur(dist, (3, 3), 0)

    overall = np.array(overall >= 0.4, dtype=np.int32)

    marker = inst_map - overall
    marker[marker < 0] = 0
    marker = ndi.binary_fill_holes(marker).astype("uint8")
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    marker = cv2.morphologyEx(marker, cv2.MORPH_OPEN, kernel)
    marker = ndi.label(marker)[0]
    marker = remove_small_objects(marker, min_size=10)

    ws_temp = segm.watershed(dist, marker, mask=inst_map)

    inst_map = remove_debris_instance(ws_temp, 18)

    return inst_map



import torch
import torch.nn as nn
import torch.nn.functional as F

from cellseg_models_pytorch.utils import filter2D, gaussian_kernel2d, tensor_one_hot


def ssim(
    img1: torch.Tensor,
    img2: torch.Tensor,
    window_size: int = 11,
    val_range: int = None,
) -> torch.Tensor:
    """Structural similarity index (SSIM).

    UNET3+ https://arxiv.org/pdf/2004.08790.pdf

    Parameters
    ----------
        img1 : torch.Tensor
            Input image 1. Shape (B, C, H, W).
        img2 : torch.Tensor
            Input image 2. Shape (B, C, H, W).
        window_size : int, default=11
            Size of the gaussian kernel.

    Returns
    -------
        torch.Tensor:
            Computed ssim loss and contrast sensitivity.
    """
    if val_range is None:
        if torch.max(img1) > 128:
            max_val = 255
        else:
            max_val = 1

        if torch.min(img1) < -0.5:
            min_val = -1
        else:
            min_val = 0
        L = max_val - min_val
    else:
        L = val_range

    (_, channel, height, width) = img1.size()
    real_size = min(window_size, height, width)
    kernel = gaussian_kernel2d(
        real_size, sigma=1.5, n_channels=channel, device=img1.device
    )

    mu1 = filter2D(img1, kernel)
    mu2 = filter2D(img2, kernel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = filter2D(img1 * img1, kernel) - mu1_sq
    sigma2_sq = filter2D(img2 * img2, kernel) - mu2_sq
    sigma12 = filter2D(img1 * img2, kernel) - mu1_mu2

    C1 = (0.01 * L) ** 2
    C2 = (0.03 * L) ** 2

    v1 = 2.0 * sigma12 + C2
    v2 = sigma1_sq + sigma2_sq + C2
    cs = torch.mean(v1 / v2)  # contrast sensitivity

    ssim_map = ((2 * mu1_mu2 + C1) * v1) / ((mu1_sq + mu2_sq + C1) * v2)

    return ssim_map, cs


class SSIM(nn.Module):
    def __init__(
        self, window_size: int = 11, return_cs: bool = False, **kwargs
    ) -> None:
        """Structural similarity index loss.

        I.e. the dissimilarity:
        (1 - SSIM(x, y)) / 2

        Parameters
        ----------
            window_size : int, default=11
                Size of the gaussian kernel.
            return_cs : bool, default=False
                Return also the the contrast sensitivity coeff.
        """
        super().__init__()
        self.window_size = window_size
        self.return_cs = return_cs

    def forward(
        self, yhat: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Compute the SSIM loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed SSIM loss (scalar).
        """
        # if target is label mask shaped (B, H, W) then one hot
        if len(target.shape) == 3:
            try:
                target = tensor_one_hot(target, n_classes=yhat.shape[1])
            except Exception:
                target = target.unsqueeze(1)

        # compute ssim
        sim, cs = ssim(yhat, target, window_size=self.window_size)
        loss = torch.clamp(1.0 - sim.mean(), min=0, max=1) / 2.0

        if self.return_cs:
            cs = torch.clamp(1.0 - cs, min=0, max=1)
            loss = (loss, cs)

        return loss


class MSSSIM(nn.Module):
    def __init__(self, window_size: int = 11, **kwargs) -> None:
        """Multi-scale SSIM loss (MS-SSIM).

        UNET3+ https://arxiv.org/pdf/2004.08790.pdf

        penalizes fuzzy boundaries

        Parameters
        ----------
            window_size : int, default=11
                Size of the gaussian kernel.
        """
        super().__init__()
        self.window_size = window_size
        self.ssim = SSIM(window_size=window_size, return_cs=True)

    def forward(
        self, yhat: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Compute the MS-SSIM loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed MS-SSIM loss (scalar).
        """
        weights = torch.tensor(
            [0.0448, 0.2856, 0.3001, 0.2363, 0.1333], device=yhat.device
        )

        levels = weights.size()[0]
        msssim = []
        mcs = []

        for _ in range(levels):
            sim, cs = self.ssim(yhat, target)
            msssim.append(sim)
            mcs.append(cs)

            yhat = F.avg_pool2d(yhat, (2, 2))
            target = F.avg_pool2d(target.float(), (2, 2))

        msssim = torch.stack(msssim)
        mcs = torch.stack(mcs)

        # Normalize to avoid NaNs during training unstable models,
        # not compliant with original definition
        msssim = (msssim + 1) / 2
        mcs = (mcs + 1) / 2

        pow1 = mcs**weights
        pow2 = msssim**weights

        loss = torch.prod(pow1[:-1] * pow2[-1])
        return loss



"""
Some of these functions adapted from the HoVer-Net repo.

https://github.com/vqdang/hover_net/blob/master/

MIT License

Copyright (c) 2020 vqdang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Dict, List, Tuple

import cv2
import numpy as np
from numba import njit, prange
from scipy import ndimage as ndi
from skimage import morphology as morph
from skimage.filters import rank
from skimage.morphology import square

__all__ = [
    "remove_small_objects",
    "binarize",
    "fix_duplicates",
    "remove_1px_boundary",
    "remap_label",
    "center_crop",
    "bounding_box",
    "get_inst_types",
    "get_inst_centroid",
    "get_type_instances",
    "one_hot",
    "type_map_flatten",
    "soft_type_flatten",
    "remove_debris_binary",
    "remove_debris_instance",
    "remove_debris_semantic",
    "fill_holes_semantic",
    "label_semantic",
    "majority_vote_parallel",
    "majority_vote_sequential",
    "med_filt_parallel",
    "med_filt_sequential",
]


# From https://github.com/scikit-image/scikit-image/blob/main/skimage/morphology/misc.py
# warning removed
def remove_small_objects(
    ar: np.ndarray,
    min_size: int = 64,
    connectivity: int = 1,
    in_place: bool = False,
    *,
    out: np.ndarray = None,
):
    """Remove objects smaller than the specified size.

    Expects ar to be an array with labeled objects, and removes objects
    smaller than min_size. If `ar` is bool, the image is first labeled.
    This leads to potentially different behavior for bool and 0-and-1
    arrays.

    Parameters
    ----------
    ar : ndarray (arbitrary shape, int or bool type)
        The array containing the objects of interest. If the array type
        is int, the ints must be non-negative.
    min_size : int, optional (default: 64)
        The smallest allowable object size.
    connectivity : int, {1, 2, ..., ar.ndim}, optional (default: 1)
        The connectivity defining the neighborhood of a pixel. Used
        during labelling if `ar` is bool.
    in_place : bool, optional (default: False)
        If ``True``, remove the objects in the input array itself.
        Otherwise, make a copy. Deprecated since version 0.19. Please
        use `out` instead.
    out : ndarray
        Array of the same shape as `ar`, into which the output is
        placed. By default, a new array is created.
    Raises
    ------
    TypeError
        If the input array is of an invalid type, such as float or
        string.
    ValueError
        If the input array contains negative values.
    Returns
    -------
    out : ndarray, same shape and type as input `ar`
        The input array with small connected components removed.
    Examples
    --------
    >>> from skimage import morphology
    >>> a = np.array([[0, 0, 0, 1, 0],
    ...               [1, 1, 1, 0, 0],
    ...               [1, 1, 1, 0, 1]], bool)
    >>> b = morphology.remove_small_objects(a, 6)
    >>> b
    array([[False, False, False, False, False],
           [ True,  True,  True, False, False],
           [ True,  True,  True, False, False]])
    >>> c = morphology.remove_small_objects(a, 7, connectivity=2)
    >>> c
    array([[False, False, False,  True, False],
           [ True,  True,  True, False, False],
           [ True,  True,  True, False, False]])
    >>> d = morphology.remove_small_objects(a, 6, out=a)
    >>> d is a
    True
    """
    if out is not None:
        in_place = False

    if in_place:
        out = ar
    elif out is None:
        out = ar.copy()

    if min_size == 0:  # shortcut for efficiency
        return out

    if out.dtype == bool:
        selem = ndi.generate_binary_structure(ar.ndim, connectivity)
        ccs = np.zeros_like(ar, dtype=np.int32)
        ndi.label(ar, selem, output=ccs)
    else:
        ccs = out

    try:
        component_sizes = np.bincount(ccs.ravel())
    except ValueError:
        raise ValueError(
            "Negative value labels are not supported. Try "
            "relabeling the input with `scipy.ndimage.label` or "
            "`skimage.morphology.label`."
        )

    too_small = component_sizes < min_size
    too_small_mask = too_small[ccs]
    out[too_small_mask] = 0

    return out


def binarize(inst_map: np.ndarray) -> np.ndarray:
    """Binarize a labelled instance map.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns
    -------
        np.ndarray:
            Binary mask. Shape (H, W). Type: uint8.
    """
    binary = np.copy(inst_map > 0)
    return binary.astype("uint8")


# ported from https://github.com/vqdang/hover_net/blob/master/src/loader/augs.py
def fix_duplicates(inst_map: np.ndarray) -> np.ndarray:
    """Re-label duplicated instances in an instance labelled mask.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns
    -------
        np.ndarray:
            The instance labelled mask without duplicated indices.
            Shape (H, W).
    """
    current_max_id = np.amax(inst_map)
    inst_list = list(np.unique(inst_map))
    inst_list.remove(0)  # 0 is background
    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.uint8)
        remapped_ids = ndi.label(inst)[0]
        remapped_ids[remapped_ids > 1] += current_max_id
        inst_map[remapped_ids > 1] = remapped_ids[remapped_ids > 1]
        current_max_id = np.amax(inst_map)

    return inst_map


# ported from https://github.com/vqdang/hover_net/blob/master/src/loader/augs.py
def remove_1px_boundary(inst_map: np.ndarray) -> np.ndarray:
    """Remove 1px around object instances.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns:
    -----------
        np.ndarray:
            The instance labelled mask with 1px of instance boundaries
            removed. Shape (H, W).
    """
    new_inst_map = np.zeros(inst_map.shape[:2], np.int32)
    inst_list = list(np.unique(inst_map))
    inst_list.remove(0)  # 0 is background
    k = morph.disk(1)

    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.uint8)
        inst = cv2.erode(inst, k, iterations=1)
        new_inst_map[inst > 0] = inst_id

    return new_inst_map


def center_crop(img: np.ndarray, ch: int, cw: int) -> np.ndarray:
    """Center crop an input image.

    Parameters
    ----------
        img : np.ndarray
            Input img. Shape (H, W).
        ch : int
            Crop height.
        cw : int
            Crop width.

    Returns
    -------
        np.ndarray:
            Center cropped image. Shape (ch, cw).
    """
    if len(img.shape) == 3:
        H, W, _ = img.shape
    else:
        H, W = img.shape

    x = W // 2 - (cw // 2)
    y = H // 2 - (ch // 2)

    if len(img.shape) == 3:
        img = img[y : y + ch, x : x + cw, :]
    else:
        img = img[y : y + ch, x : x + cw]

    return img


# Ported from https://github.com/vqdang/hover_net/blob/master/src/misc/utils.py
def bounding_box(inst_map: np.ndarray) -> List[int]:
    """Bounding box coordinates for an instance that is given as input.

    This assumes that the `inst_map` has only one instance in it.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns
    -------
        List[int]:
            List of the origin- and end-point coordinates of the bbox.
    """
    rows = np.any(inst_map, axis=1)
    cols = np.any(inst_map, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    # due to python indexing, need to add 1 to max
    # else accessing will be 1px in the box, not out
    rmax += 1
    cmax += 1

    return [rmin, rmax, cmin, cmax]


# ported from https://github.com/vqdang/hover_net/tree/master/src/metrics/sample
def remap_label(inst_map: np.ndarray) -> np.ndarray:
    """Rename all instance id so that the id is contiguous.

    I.e [0, 1, 2, 3] not [0, 2, 4, 6].

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns
    -------
        np.ndarray:
            Instance labelled mask with remapped contiguous labels.
    """
    inst_list = list(np.unique(inst_map))
    inst_list.remove(0)
    if len(inst_list) == 0:
        return inst_map  # no label

    new_inst_map = np.zeros(inst_map.shape, np.int32)
    for idx, inst_id in enumerate(inst_list):
        new_inst_map[inst_map == inst_id] = idx + 1

    return new_inst_map


# Ported from https://github.com/vqdang/hover_net/blob/master/src/misc/utils.py
def get_inst_centroid(inst_map: np.ndarray) -> np.ndarray:
    """Get centroid x, y coordinates from each unique nuclei instance.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).

    Returns:
    ----------
        np.ndarray:
            An array of shape (num_instances, 2).
    """
    inst_centroid_list = []
    inst_id_list = list(np.unique(inst_map))
    for inst_id in inst_id_list[1:]:  # avoid 0 i.e background
        mask = np.array(inst_map == inst_id, np.uint8)
        inst_moment = cv2.moments(mask)
        inst_centroid = [
            (inst_moment["m10"] / inst_moment["m00"]),
            (inst_moment["m01"] / inst_moment["m00"]),
        ]
        inst_centroid_list.append(inst_centroid)

    return np.array(inst_centroid_list)


def get_inst_types(inst_map: np.ndarray, type_map: np.ndarray) -> np.ndarray:
    """Get the each instance type of a instance labelled mask.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        type_map : np.ndarray
            Type labelled mask. Shape (H, W).

    Returns
    -------
        An np.ndarray of shape (num_instances, 1).
    """
    inst_ids = list(np.unique(inst_map))

    if 0 in inst_ids:
        inst_ids.remove(0)

    inst_types = np.full((len(inst_ids), 1), 0, dtype=np.int32)
    for j, id_ in enumerate(inst_ids):
        inst_type = np.unique(type_map[inst_map == id_])[0]
        inst_types[j] = inst_type

    return inst_types


def get_type_instances(
    inst_map: np.ndarray, type_map: np.ndarray, class_num: int
) -> np.ndarray:
    """Get the instances of the input that belong to class `class_num`.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        type_map : np.ndarray
            Type labelled mask. Shape (H, W).
        class_num : int
            Class label.

    Returns
    -------
        np.ndarray:
            An array  of shape (H, W) where the values equalling
            `class_num` are dropped.
    """
    t = type_map.astype("uint8") == class_num
    imap = np.copy(inst_map)
    imap[~t] = 0

    return imap


def one_hot(type_map: np.ndarray, num_classes: int) -> np.ndarray:
    """Convert a type labelled mask of shape (H, W) to one hot (H, W, C).

    Parameters
    ----------
        type_map : np.ndarray
            Type labelled mask. Shape (H, W).
        num_classes (int):
            Number of classes in the dataset.

    Raises:
    -------
        ValueError:
            If the given `num_classes` is less than the observed number
            of classes

    Returns:
    -----------
        np.ndarray:
            An array of the input array (H, W) in one hot format.
            Shape: (H, W, num_classes). Dtype: float64
    """
    ntypes = len(np.unique(type_map)) - 1
    if num_classes < ntypes:
        raise ValueError(
            f"""`num_classes` is less than observed number of classes.
            The type map contains {ntypes}. Got `num_classes`: {num_classes}"""
        )

    return np.eye(num_classes + 1)[type_map]


def type_map_flatten(type_map: np.ndarray) -> np.ndarray:
    """Inverted one-hot.

    Converts a one hot type map of shape (H, W, C) to a single channel
    indice map of shape (H, W).

    Parameters
    ----------
        type_map : np.ndarray
            Type labelled mask. Shape (H, W, C).

    Returns
    -------
        np.ndarray:
            Flattened one hot np.ndarray. Shape (H, W).
    """
    type_out = np.zeros([type_map.shape[0], type_map.shape[1]])
    for t in range(type_map.shape[-1]):
        type_out += type_map[..., t] * t

    return type_out


def soft_type_flatten(type_map: np.ndarray) -> np.ndarray:
    """Flatten a one hot soft mask of shape (H, W, C).

    Parameters
    ----------
        type_map : np.ndarray
            Type labelled mask. Shape (H, W, C).

    Returns
    -------
        np.ndarray:
            Flattened soft mask. Shape (H, W).
    """
    type_out = np.zeros([type_map.shape[0], type_map.shape[1]])
    for i in range(1, type_map.shape[-1]):
        type_tmp = type_map[..., i]
        type_out += type_tmp

    return type_out


def remove_debris_binary(binary_mask: np.ndarray) -> np.ndarray:
    """Take in a binary mask -> fill holes -> removes small objects.

    Parameters
    -----------
        binary_mask : np.ndarray
            A binary mask. Shape (H, W)|(H, W, C).

    Returns
    -------
        np.ndarray:
            Cleaned binary mask of shape (H, W).
    """
    if len(binary_mask.shape) == 3:
        binary_mask = binary_mask[..., 1]

    out_mask = ndi.binary_fill_holes(binary_mask)
    out_mask = remove_small_objects(binary_mask.astype(bool), min_size=10)

    return out_mask.astype("u1")


def remove_debris_instance(inst_map: np.ndarray, min_size: int = 10):
    """Remove small objects from an inst map.

    (When skimage and ndimage fails)

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        min_size : int, default=10
            Min size for the objects that are left untouched.

    Returns
    -------
        np.ndarray:
            Cleaned instance labelled mask of shape (H, W).
    """
    res = np.zeros(inst_map.shape, np.int32)
    for ix in np.unique(inst_map)[1:]:
        nuc_map = np.copy(inst_map == ix)

        y1, y2, x1, x2 = bounding_box(nuc_map)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= inst_map.shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= inst_map.shape[0] - 1 else y2
        nuc_map_crop = nuc_map[y1:y2, x1:x2].astype("i4")

        nuc_map_crop = remove_small_objects(
            nuc_map_crop.astype(bool), min_size, connectivity=1
        ).astype("i4")

        nuc_map_crop[nuc_map_crop > 0] = ix
        res[y1:y2, x1:x2] += nuc_map_crop

    return res


def remove_debris_semantic(sem_map: np.ndarray, min_size: int = 10000):
    """Remove small objects from a semantic area map.

    Parameters
    ----------
        sem_map : np.ndarray
            Semantic segmentation mask. Shape (H, W).
        min_size : int, default=10000
            Min size for the objects that are left untouched

    Returns
    -------
        np.ndarray:
            Cleaned semantic segmentation mask of shape (H, W).
    """
    res = np.copy(sem_map)
    classes = np.unique(sem_map)

    # skip bg
    if 0 in classes:
        classes = classes[1:]

    for i in classes:

        area = np.array(res == i, np.int32)
        inst_map = ndi.label(area)[0]
        labels, counts = np.unique(inst_map, return_counts=True)

        for label, npixls in zip(labels, counts):

            if npixls < min_size:
                res[inst_map == label] = 0

                # get the fill label
                y1, y2, x1, x2 = bounding_box(inst_map == label)
                y1 = y1 - 2 if y1 - 2 >= 0 else y1
                x1 = x1 - 2 if x1 - 2 >= 0 else x1
                x2 = x2 + 2 if x2 + 2 <= res.shape[1] - 1 else x2
                y2 = y2 + 2 if y2 + 2 <= res.shape[0] - 1 else y2
                labels, counts = np.unique(res[y1:y2, x1:x2], return_counts=True)

                if 0 in labels and len(labels) > 1:
                    labels = labels[1:]
                    counts = counts[1:]

                fill_label = labels[np.argmax(counts)]
                res[inst_map == label] = fill_label

    return res


def fill_holes_semantic(sem_map: np.ndarray, min_size: int = 5000):
    """Fill holes (background) from a semantic segmentation map.

    Parameters
    ----------
        sem_map : np.ndarray
            Semantic segmentation mask. Shape (H, W).
        min_size : int, default=5000
            Min size for the objects that are left untouched.

    Returns
    -------
        np.ndarray:
            Cleaned semantic segmentation mask of shape (H, W).
    """
    res = np.copy(sem_map)
    bg = res == 0
    bg_objs = ndi.label(bg)[0]

    for i in np.unique(bg_objs)[1:]:
        y1, y2, x1, x2 = bounding_box(bg_objs == i)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= res.shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= res.shape[0] - 1 else y2
        crop = res[y1:y2, x1:x2]

        labels, counts = np.unique(crop, return_counts=True)

        if counts[0] > min_size:
            continue

        if len(counts) == 1:
            continue

        # skip 0 index
        labels = labels[1:]
        counts = counts[1:]

        # fill bg objs
        fill_label = labels[np.argmax(counts)]
        crop[crop == 0] = fill_label
        res[y1:y2, x1:x2] = crop

    return res


def label_semantic(sem_map: np.ndarray, sort: bool = True) -> np.ndarray:
    """Labels a given semantic segmentation map.

    Parameters
    ----------
        sem_map : np.ndarray
            Semantic segmentation map. Shape (H, W)
        sort : bool, default=True
            Sort the semantic areas by size in descending order.

    Returns
    -------
        np.ndarray:
            The labelled segmentation map. Shape (H, W).
    """
    sem_inst = np.zeros_like(sem_map)

    counter = 0
    classes, counts = np.unique(sem_map, return_counts=True)

    if 0 in classes:
        classes = classes[1:]
        counts = counts[1:]

    if sort:
        classes = classes[np.argsort(-counts)]

    for c in classes:
        obj_insts = ndi.label(sem_map == c)[0]
        labels = np.unique(obj_insts)

        # If all pixels belong to same (non bg) class, dont slice
        if len(labels) > 1:
            labels = labels[1:]

        for label in labels:
            counter += 1
            obj = obj_insts == label
            sem_inst[obj] += counter

    return sem_inst


NUM_COLORS = {
    0: (255.0, 0.0, 55.0),
    1: (255.0, 0.0, 0.0),
    2: (0.0, 200.0, 100.0),
    3: (220.0, 220.0, 55.0),
    4: (0.0, 110.0, 155.0),
    5: (50.0, 50.0, 155.0),
    6: (220.0, 220.0, 55.0),
    7: (200.0, 50.0, 55.0),
    8: (155.0, 110.0, 155.0),
}


def draw_thing_contours(
    inst_map: np.ndarray,
    image: np.ndarray,
    type_map: np.ndarray = None,
    thickness: int = 2,
    fill_contours: bool = False,
    classes: Dict[str, int] = None,
    colors: Dict[str, Tuple[float, float, float]] = None,
) -> np.ndarray:
    """Find coloured contours for an instance labelled mask.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance segmentation map. Shape (H, W).
        image : np.ndarray
            Original image.
        type_map : np.ndarray, optional
            Semantic segmentation map. Shape (H, W)
        fill_contours bool, default=False
            If True, contours are filled
        thickness : int, default=2
            Thickness of the contour lines
        classes : Dict[str, int], optional
            Classes dict e.g. {"bg":0, "cancer":1, "immune":2}
        colors : Dict[str, Tuple[float, float, float]], optional
            Color dict for the classes.
            E.g. {"cancer": (125., 100. ,122.), "immune": (56., 37, 160.)}

    Returns
    -------
        np.ndarray:
            The contours overlaid on top of original image. Shape: (H, W, 3).
    """
    bg = np.copy(image)

    shape = inst_map.shape[:2]
    nuc_list = list(np.unique(inst_map))

    if 0 in nuc_list:
        nuc_list.remove(0)  # 0 is background

    for _, nuc_id in enumerate(nuc_list):
        inst = np.array(inst_map == nuc_id, np.uint8)

        y1, y2, x1, x2 = bounding_box(inst)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= shape[0] - 1 else y2

        img_crop = image[y1:y2, x1:x2]
        inst_crop = inst[y1:y2, x1:x2]
        inst_bg_crop = bg[y1:y2, x1:x2]
        contours = cv2.findContours(inst_crop, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[
            0
        ]

        inst_color = None
        if classes is not None:
            class_num = np.unique(type_map[inst > 0].astype("uint8"))[0]
            class_name = [key for key, item in classes.items() if class_num == item][0]

            if colors is not None:
                if class_name in colors.keys():
                    inst_color = colors[class_name]
            else:
                inst_color = NUM_COLORS[class_num]
        elif type_map is not None:
            type_crop = type_map[y1:y2, x1:x2]
            type = np.unique(type_crop[inst_crop > 0])[0]
            inst_color = NUM_COLORS[type]

        if fill_contours:
            cv2.drawContours(
                inst_bg_crop,
                [max(contours, key=cv2.contourArea)],
                contourIdx=-1,
                color=inst_color,
                thickness=-1,
            )

            # blend with original image
            alpha = 0.50
            inst_bg_crop = cv2.addWeighted(img_crop, 1 - alpha, inst_bg_crop, alpha, 0)
        else:
            cv2.drawContours(
                inst_bg_crop,
                contours,
                contourIdx=-1,
                color=inst_color,
                thickness=thickness,
            )

        bg[y1:y2, x1:x2] = inst_bg_crop

    return bg


def draw_stuff_contours(
    inst_map: np.ndarray,
    image: np.ndarray,
    type_map: np.ndarray,
    fill_contours: bool = False,
    thickness: int = 2,
    classes: Dict[str, int] = None,
    colors: Dict[str, Tuple[float, float, float]] = None,
) -> np.ndarray:
    """Find coloured contours for a semantic segmentation mask.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance segmentation map. Shape: (H, W).
        image : np.ndarray
            Original image. Shape: (H, W, 3).
        type_map : np.ndarray, optional
            Semantic segmentation map. Shape (H, W)
        fill_contours bool, default=False
            If True, contours are filled
        thickness : int, default=2
            Thickness of the contour lines
        classes : Dict[str, int], optional
            Classes dict e.g. {"bg":0, "cancer":1, "immune":2}
        colors : Dict[str, Tuple[float, float, float]], optional
            Color dict for the classes.
            E.g. {"cancer": (125., 100. ,122.), "immune": (56., 37, 160.)}

    Returns:
    ---------
        np.ndarray:
            The contours overlaid on top of original image. Shape (H, W, 3).
    """
    bg = np.copy(image)

    obj_list = list(np.unique(inst_map))

    if 0 in obj_list:
        obj_list.remove(0)  # 0 is background

    for obj in obj_list:
        inst = np.array(inst_map == obj, np.uint8)
        contours = cv2.findContours(inst, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[0]

        inst_color = None

        if classes is not None:
            class_num = np.unique(type_map[inst > 0].astype("uint8"))[0]
            class_name = [key for key, item in classes.items() if class_num == item][0]
            if colors is not None:
                if class_name in colors.keys():
                    inst_color = colors[class_name]
            else:
                if type_map is not None:
                    inst_color = NUM_COLORS[class_num]
        elif type_map is not None:
            type = np.unique(type_map[inst > 0])[0]
            inst_color = NUM_COLORS[type]

        if fill_contours:
            cv2.drawContours(
                bg,
                [max(contours, key=cv2.contourArea)],
                contourIdx=-1,
                color=inst_color,
                thickness=-1,
            )
        else:
            cv2.drawContours(
                bg, contours, contourIdx=-1, color=inst_color, thickness=thickness
            )

        # blend with original image
        alpha = 0.8
        bg = cv2.addWeighted(image, 1 - alpha, bg, alpha, 0)

    return bg


@njit(parallel=True)
def majority_vote_parallel(type_map: np.ndarray, inst_map: np.ndarray) -> np.ndarray:
    """Do a majority voting on the type_map pixels in parallel.

    I.e. Given a raw argmaxed type_map and an instance labelled mask, for each instance,
    decide the class of the instance with majority voting on the corresponding type_map
    pixels.

    NOTE: This gives a performance boost over the sequential version.

    Parameters
    ----------
        type_map : np.ndarray
            A raw type map (after an argmax function). Shape: (H, W).
        inst_map : np.ndarray
            An instance labelled mask. Shape: (H, W).

    Returns
    -------
        np.ndarray:
            Post-processed type map. Shape: (H, W).
    """
    tmap = np.zeros_like(inst_map)
    types = np.unique(type_map)
    ntypes = np.max(types)
    pred_id_list = np.unique(inst_map)[1:]
    y, x = np.nonzero(inst_map)

    for ix in prange(len(pred_id_list)):
        inst_id = pred_id_list[ix]
        type_pxls = np.zeros(ntypes + 1)

        for j, i in zip(y, x):
            if inst_map[j, i] == inst_id:
                pxl_type = type_map[j, i]
                type_pxls[int(pxl_type)] += 1

        for j, i in zip(y, x):
            if inst_map[j, i] == inst_id:
                tmap[j, i] = np.argmax(type_pxls)

    return tmap


def majority_vote_sequential(
    type_map: np.ndarray,
    inst_map: np.ndarray,
) -> np.ndarray:
    """Do a majority voting on the type_map pixels by broadcasting.

    Adapted from:
    https://github.com/vqdang/hover_net/blob/master/models/hovernet/post_proc.py

    Parameters
    ----------
        type_map : np.ndarray
            A raw type map (after an argmax function). Shape: (H, W).
        inst_map : np.ndarray
            An instance labelled mask. Shape: (H, W).

    Returns
    -------
        np.ndarray:
            Post-processed type map. Shape: (H, W).
    """
    type_map = binarize(inst_map) * type_map
    pred_id_list = np.unique(inst_map)[1:]
    for inst_id in pred_id_list:
        inst_type = type_map[inst_map == inst_id]
        type_list, type_pixels = np.unique(inst_type, return_counts=True)
        type_list = list(zip(type_list, type_pixels))
        type_list = sorted(type_list, key=lambda x: x[1], reverse=True)
        cell_type = type_list[0][0]

        if cell_type == 0:
            if len(type_list) > 1:
                cell_type = type_list[1][0]

        type_map[inst_map == inst_id] = cell_type

    return type_map


@njit(parallel=True)
def med_filt_parallel(sem_map: np.ndarray, kernel_size: Tuple[int, int]) -> np.ndarray:
    """Parallel median filter.

    NOTE: Can bring latency benefits if a lot of cores are available. If not,
    only slows down things.

    Parameters
    ----------
        sem_map : np.ndarray
            Input semantic segmentation prob map. Shape: (C, H, W). Dtype: uint8.
        kernel_size : Tuple[int, int]
            Size of the kernel.

    Returns
    -------
        Median filtered probability map. Shape: (C, H, W). Dtype: uint8.
    """
    _, ny, nx = sem_map.shape
    ky = kernel_size[0] // 2
    kx = kernel_size[1] // 2
    z, y, x = np.nonzero(sem_map)

    smap = np.zeros_like(sem_map)
    coords = list(zip(z, y, x))

    for coord_ix in prange(len(coords)):
        c, j, i = coords[coord_ix]
        j_start = max(j - ky, 0)
        j_end = min(j + ky, ny - 1)
        i_start = max(i - kx, 0)
        i_end = min(i + kx, nx - 1)

        box = sem_map[c, j_start : j_end + 1, i_start : i_end + 1]
        smap[c, j, i] = np.median(box)

    return smap


def med_filt_sequential(sem_map: np.ndarray, kernel_width: int = 15) -> np.ndarray:
    """Sequential median filter.

    Parameters
    ----------
        sem_map : np.ndarray
            Input semantic segmentation prob map. Shape: (C, H, W). Dtype: uint8.

    Returns
    -------
        Median filtered probability map. Shape: (C, H, W). Dtype: uint8.
    """
    sem = np.zeros_like(sem_map)
    for i in range(sem_map.shape[0]):
        sem[i] = rank.median(sem_map[i], footprint=square(kernel_width))

    return sem



import shutil
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
from tqdm import tqdm

from ..utils import FileHandler, fix_duplicates

try:
    from ..datasets import SegmentationFolderDataset
    from ._basemodule import BaseDataModule
    from .downloader import SimpleDownloader
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the PannukeDataModule, requests, pytorch-lightning, & albumentations "
        "libraries are needed. Install with "
        "`pip install requests pytorch-lightning albumentations`"
    )


class PannukeDataModule(BaseDataModule):
    def __init__(
        self,
        save_dir: str,
        fold_split: Dict[str, int],
        img_transforms: List[str],
        inst_transforms: List[str],
        normalization: str = None,
        batch_size: int = 8,
        num_workers: int = 8,
        **kwargs,
    ) -> None:
        """Set up pannuke datamodule..

        The patches will be saved in directories:
        - `{save_dir}/train/*`
        - `{save_dir}/test/*`
        - `{save_dir}/valid/*`

        References
        ----------
        Gamper, J., Koohbanani, N., Benet, K., Khuram, A., & Rajpoot, N. (2019)
        PanNuke: an open pan-cancer histology dataset for nuclei instance segmentation
        and classification. In European Congress on Digital Pathology (pp. 11-19).

        Gamper, J., Koohbanani, N., Graham, S., Jahanifar, M., Khurram, S., Azam, A.,
        Hewitt, K., & Rajpoot, N. (2020). PanNuke Dataset Extension, Insights and
        Baselines. arXiv preprint arXiv:2003.10778.

        License: https://creativecommons.org/licenses/by-nc-sa/4.0/

        Parameters
        ----------
            save_dir : str
                Path to directory where the pannuke data will be saved.
            fold_split : Dict[str, int]
                Defines how the folds are split into train, valid, and test sets.
                E.g. {"train": 1, "valid": 2, "test": 3}
            img_transforms : List[str]
                A list containing all the transformations that are applied to the input
                images and corresponding masks. Allowed ones: "blur", "non_spatial",
                "non_rigid", "rigid", "hue_sat", "random_crop", "center_crop", "resize"
            inst_transforms : List[str]
                A list containg all the transformations that are applied to only the
                instance labelled masks. Allowed ones: "cellpose", "contour", "dist",
                "edgeweight", "hovernet", "omnipose", "smooth_dist", "binarize"
            normalization : str, optional
                Apply img normalization after all the transformations. One of "minmax",
                "norm", "percentile", None.
            batch_size : int, default=8
                Batch size for the dataloader.
            num_workers : int, default=8
                number of cpu cores/threads used in the dataloading process.

        Example
        -------
            >>> from pathlib import Path
            >>> from cellseg_models_pytorch.datamodules import PannukeDataModule

            >>> fold_split = {"train": 1, "valid": 2, "test": 3}
            >>> save_dir = Path.home() / "pannuke"
            >>> pannuke_module = PannukeDataModule(
                    save_dir=save_dir,
                    fold_split=fold_split,
                    inst_transforms=["dist", "stardist"],
                    img_transforms=["blur", "hue_sat"],
                    normalization="percentile",
                    n_rays=32
                )

            >>> # pannuke_module.download(save_dir) # just the downloading
            >>> pannuke_module.prepare_data() # download and process the pannuke data
        """
        super().__init__(batch_size, num_workers)
        allowed = ("train", "valid", "test")
        if not all([k in allowed for k in fold_split.keys()]):
            raise ValueError(
                f"`fold_split` keys need to be in {allowed}. "
                f"Got: {list(fold_split.keys())}"
            )

        self.save_dir = Path(save_dir)
        self.fold_split = fold_split
        self.img_transforms = img_transforms
        self.inst_transforms = inst_transforms
        self.normalization = normalization
        self.kwargs = kwargs if kwargs is not None else {}

    @property
    def type_classes(self) -> Dict[str, int]:
        """Pannuke cell type classes."""
        return {
            "bg": 0,
            "neoplastic": 1,
            "inflammatory": 2,
            "connective": 3,
            "dead": 4,
            "epithelial": 5,
        }

    @staticmethod
    def download(root: str) -> None:
        """Download the pannuke dataset from online."""
        for ix in [1, 2, 3]:
            url = f"https://warwick.ac.uk/fac/cross_fac/tia/data/pannuke/fold_{ix}.zip"
            SimpleDownloader.download(url, root)
        PannukeDataModule.extract_zips(root, rm=True)

    def prepare_data(self, rm_orig: bool = True) -> None:
        """Prepare the pannuke datasets.

        1. Download pannuke folds from:
            "https://warwick.ac.uk/fac/cross_fac/tia/data/pannuke/"
        2. Pre-process and split the images and masks into train, valid and test sets.

        Parameters
        ----------
            rm_orig : bool, default=False
                After processing all the files, If True, removes the original
                un-processed files.
        """
        folds_found = [
            d.name
            for d in self.save_dir.iterdir()
            if "fold" in d.name.lower() and d.is_dir()
        ]
        phases_found = [
            d.name
            for d in self.save_dir.iterdir()
            if d.name in ("train", "test", "valid") and d.is_dir()
        ]

        if len(folds_found) < 3 and not phases_found:
            print(f"Found {len(folds_found)} folds. " "Downloading all three...")
            PannukeDataModule.download(self.save_dir)
        else:
            print("Found all folds. Skip downloading.")

        if not phases_found:
            print("Processing files...")
            fold_paths = self._get_fold_paths(self.save_dir)

            for phase, fold_ix in self.fold_split.items():
                save_im_dir = self.save_dir / phase / "images"
                save_mask_dir = self.save_dir / phase / "labels"
                self._process_pannuke_fold(
                    fold_paths, save_im_dir, save_mask_dir, fold_ix, phase
                )
        else:
            print(
                "Found processed pannuke data. "
                "If in need of a re-download, please empty the `save_dir` folder."
            )

        if rm_orig:
            for d in self.save_dir.iterdir():
                if "fold" in d.name.lower():
                    shutil.rmtree(d)

    def setup(self, stage: Optional[str] = None) -> None:
        """Set up the train, valid, and test datasets."""
        self.trainset = SegmentationFolderDataset(
            path=self.save_dir / "train" / "images",
            mask_path=self.save_dir / "train" / "labels",
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

        self.validset = SegmentationFolderDataset(
            path=self.save_dir / "valid" / "images",
            mask_path=self.save_dir / "valid" / "labels",
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

        self.testset = SegmentationFolderDataset(
            path=self.save_dir / "test" / "images",
            mask_path=self.save_dir / "test" / "labels",
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

    def _get_fold_paths(self, path: Path) -> Dict[str, Path]:
        """Get the paths to the .npy files in all of the fold folders."""
        return {
            f"{file.parts[-2]}_{file.name[:-4]}": file
            for dir1 in path.iterdir()
            if dir1.is_dir()
            for dir2 in dir1.iterdir()
            if dir2.is_dir()
            for dir3 in dir2.iterdir()
            if dir3.is_dir()
            for file in dir3.iterdir()
            if file.is_file() and file.suffix == ".npy"
        }

    def _process_pannuke_fold(
        self,
        fold_paths: Dict[str, Path],
        save_im_dir: Path,
        save_mask_dir: Path,
        fold: int,
        phase: str,
    ) -> None:
        """Save the pannuke patches .mat files in 'train', 'valid' & 'test' folders."""
        # Create directories for the files.
        Path(save_im_dir).mkdir(parents=True, exist_ok=True)
        Path(save_mask_dir).mkdir(parents=True, exist_ok=True)

        masks = np.load(fold_paths[f"fold{fold}_masks"]).astype("int32")
        imgs = np.load(fold_paths[f"fold{fold}_images"]).astype("uint8")
        types = np.load(fold_paths[f"fold{fold}_types"])

        with tqdm(total=len(types)) as pbar:
            pbar.set_description(f"fold{fold}/{phase}")
            for tissue_type in np.unique(types):
                imgs_by_type = imgs[types == tissue_type]
                masks_by_type = masks[types == tissue_type]
                for j in range(imgs_by_type.shape[0]):
                    name = f"{tissue_type}_fold{fold}_{j}"
                    fn_im = Path(save_im_dir / name).with_suffix(".png")
                    FileHandler.write_img(fn_im, imgs_by_type[j, ...])

                    temp_mask = masks_by_type[j, ...]
                    type_map = self._get_type_map(temp_mask)
                    inst_map = self._get_inst_map(temp_mask[..., 0:5])

                    fn_mask = Path(save_mask_dir / name).with_suffix(".mat")
                    FileHandler.write_mask(fn_mask, inst_map, type_map)
                    pbar.update(1)

    def _get_type_map(self, pannuke_mask: np.ndarray) -> np.ndarray:
        """Convert the pannuke mask to type map of shape (H, W)."""
        mask = pannuke_mask.copy()
        mask[mask > 0] = 1

        # init type_map and set the background channel
        # of the pannuke mask as the first channel
        type_map = np.zeros_like(mask)
        type_map[..., 0] = mask[..., -1]
        for i, j in enumerate(range(1, mask.shape[-1])):
            type_map[..., j] = mask[..., i]

        return np.argmax(type_map, axis=-1)

    def _get_inst_map(self, pannuke_mask: np.ndarray) -> np.ndarray:
        """Convert pannuke mask to inst_map of shape (H, W)."""
        mask = pannuke_mask.copy()

        inst_map = np.zeros(mask.shape[:2], dtype="int32")
        for i in range(mask.shape[-1]):

            insts = mask[..., i]
            inst_ids = np.unique(insts)[1:]
            for inst_id in inst_ids:
                inst = np.array(insts == inst_id, np.uint8)
                inst_map[inst > 0] += inst_id

        # fix duplicated instances
        inst_map = fix_duplicates(inst_map)
        return inst_map
